import dash_html_components as html
import dash_core_components as dcc
import dash_table

from chemicalc.reference_spectra import reference_stars
from chemicalc.s2n import keck_options
from .utils import spectrographs, prior_tooltips, snr_options, etc_options
from .utils import sample_labels, lab_tab_text

import numpy as np
from plotly.subplots import make_subplots
import plotly.graph_objs as go

UC_colors = {'blue': 'rgb(0, 50, 98)',
             'white': 'rgb(255, 255, 255)'}

n_combo = 3
n_inst = 3


def create_instrument_selection(combo_no, inst_no):
    inst_presets = html.Div([
        dcc.Dropdown(id=f'preset-{combo_no}-{inst_no}',
                     placeholder='Load from presets',
                     options=[{'label': spec_name, 'value': spec_name} for spec_name in spectrographs.keys()],
                     style={'marginRight': 10})
    ])
    inst_name = html.Div([
        html.Label('Name: ',
                   style={'float': 'Left',
                          'padding': '10px 0px',
                          'display': 'tableCell',
                          'verticalAlign': 'Bottom',
                          'marginTop': 3}),
        dcc.Input(id=f'instname-{combo_no}-{inst_no}',
                  placeholder='-',
                  value='',
                  style={'width': '75%',
                         'float': 'Right',
                         'textAlign': 'Right',
                         'display': 'tableCell',
                         'verticalAlign': 'Bottom',
                         'marginTop': 3,
                         'marginRight': 10}
    )
        ], style={'display': 'block'})
    inst_wave = html.Div([
        html.Label(u'Wavelength (\u212B):',
                   style={'float': 'left',
                          'padding': '10px 10px 0px 0px'}),
        html.Div([
            dcc.Input(id=f'wavemin-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      step=10, placeholder="start",
                      style={'width': 75,
                             'display': 'inline-block'}),
            html.P(u'Min: 3000 \u212B',
                   style={'fontStyle': 'italic',
                          'fontSize': 12,
                          'padding': '0 0'})
        ], style={'display': 'inline-block',
                  'float': 'Left'}),
        html.Div(html.Label('-'),
                 style={'float': 'Left',
                        'padding': '10px 0px',
                        'display': 'inline_block'}),
        html.Div([
            dcc.Input(id=f'wavemax-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      step=10, placeholder="end",
                      style={'width': 75,
                             'display': 'inline-block'}),
            html.P(u'Max: 18000 \u212B',
                   style={'fontStyle': 'italic',
                          'fontSize': 12})
        ], style={'display': 'inline-block',
                  'float': 'Left',
                  'marginRight': 10})
    ])
    inst_res = html.Div([
        html.Label(u'Res. Power (\u0394\u03BB/\u03BB):',
                   style={'float': 'Left',
                          'verticalAlign': 'Middle',
                          'padding': '5px 0px'}),
        dcc.Input(id=f'res-{combo_no}-{inst_no}',
                  type='number', inputMode='numeric',
                  placeholder='-',
                  min=1e2, max=1e5, step=1e2,
                  style={'width': '40%',
                         'float': 'Right',
                         'marginTop': -3,
                         'marginRight': 10,
                         'textAlign': 'Right'}),
    ], style={'display': 'block'})
    inst_samp = html.Div([
        html.Label(r'Pixels / FWHM:',
                   style={'float': 'Left',
                          'verticalAlign': 'Middle',
                          'padding': '5px 0px'}),
        dcc.Input(id=f'samp-{combo_no}-{inst_no}',
                  type='number', inputMode='numeric',
                  placeholder='-',
                  min=1, max=10, step=1,
                  style={'width': '40%',
                         'float': 'Right',
                         'marginTop': 3,
                         'marginRight': 10,
                         'textAlign': 'Right'}),
    ], style={'display': 'block'})
    inst_snr = html.Details([
        html.Summary('SNR Configuration', style={'textAlign': 'Left'}),
        dcc.Dropdown(id=f'snr-type-{combo_no}-{inst_no}',
                     placeholder='Select SNR type',
                     options=[{'label': opt, 'value': opt} for opt in snr_options],
                     value=snr_options[0],
                     style={'textAlign': 'Left',
                            'marginTop': 5,
                            'marginRight': 10,}),
        html.Div(id=f'constant-snr-{combo_no}-{inst_no}', children=[
            html.Label(r'SNR / pixel:',
                       style={'float': 'Left',
                              'verticalAlign': 'Middle',
                              'padding': '10px 0px'}),
            dcc.Input(id=f'snr-const-value-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      placeholder='-',
                      value=100,
                      min=1, max=1000, step=10,
                      style={'width': '40%',
                             'float': 'Right',
                             'marginTop': 3,
                             'marginRight': 10,
                             'textAlign': 'Right'})
        ]),
        dcc.Dropdown(id=f'etc-dropdown-{combo_no}-{inst_no}',
                     placeholder='Choose ETC...',
                     options=[{'label': opt, 'value': opt} for opt in etc_options],
                     clearable=True,
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-inst-{combo_no}-{inst_no}',
                     placeholder='Select Instrument...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['instrument']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),

        dcc.Dropdown(id=f'wmko-grating-{combo_no}-{inst_no}',
                     placeholder='Select Grating...',
                     options=[],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-grism-{combo_no}-{inst_no}',
                     placeholder='Select Grism...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['grism (LRIS)']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-slitwidth-{combo_no}-{inst_no}',
                     placeholder='Select Slitwidth...',
                     options=[],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-dichroic-{combo_no}-{inst_no}',
                     placeholder='Select Dichroic...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['dichroic (LRIS)']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-binning-{combo_no}-{inst_no}',
                     placeholder='Select Binning...',
                     options=[],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-cwave-{combo_no}-{inst_no}',
                     placeholder='Select Central Wavelength...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['central wavelength (DEIMOS)']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),

        html.Div(id=f'wmko-exptime-{combo_no}-{inst_no}', children=[
            html.Label(r'Exposure Time (s):',
                       style={'float': 'Left',
                              'verticalAlign': 'Middle',
                              'padding': '10px 0px'}),
            dcc.Input(id=f'wmko-exptime-value-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      placeholder='-',
                      value=3600,
                      min=1, step=1000,
                      style={'width': '40%',
                             'float': 'Right',
                             'marginTop': 3,
                             'marginRight': 10,
                             'textAlign': 'Right'})
        ], style={'display': 'None'}),
        html.Div(style={'clear': 'both'}),
        html.Div(id=f'wmko-mag-{combo_no}-{inst_no}', children=[
            html.Label(r'Apparent Magnitude:',
                       style={'float': 'Left',
                              'verticalAlign': 'Middle',
                              'padding': '10px 0px'}),
            dcc.Input(id=f'wmko-mag-value-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      placeholder='-',
                      value=19,
                      min=0, max=30, step=0.5,
                      style={'width': '40%',
                             'float': 'Right',
                             'marginTop': 3,
                             'marginRight': 10,
                             'textAlign': 'Right'})
        ], style={'display': 'None'}),
        html.Div(style={'clear': 'both'}),
        dcc.Dropdown(id=f'wmko-magtype-{combo_no}-{inst_no}',
                     placeholder='AB or Vega...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['mag type']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-filter-{combo_no}-{inst_no}',
                     placeholder='Select Filter...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['filter']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),
        dcc.Dropdown(id=f'wmko-star-{combo_no}-{inst_no}',
                     placeholder='Select Stellar Template...',
                     options=[{'label': opt, 'value': opt} for opt in keck_options['template']],
                     style={'textAlign': 'Left',
                            'marginRight': 10,
                            'display': 'None'}),

        html.Div(style={'clear': 'both'}),
        html.Div(id=f'wmko-airmass-{combo_no}-{inst_no}', children=[
            html.Label(r'Airmass:',
                       style={'float': 'Left',
                              'verticalAlign': 'Middle',
                              'padding': '10px 0px'}),
            dcc.Input(id=f'wmko-airmass-value-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      placeholder='-',
                      value=1.1,
                      min=1, max=5, step=0.1,
                      style={'width': '40%',
                             'float': 'Right',
                             'marginTop': 3,
                             'marginRight': 10,
                             'textAlign': 'Right'})
        ], style={'display': 'None'}),
        html.Div(style={'clear': 'both'}),
        html.Div(id=f'wmko-seeing-{combo_no}-{inst_no}', children=[
            html.Label(r'Seeing ("):',
                       style={'float': 'Left',
                              'verticalAlign': 'Middle',
                              'padding': '10px 0px'}),
            dcc.Input(id=f'wmko-seeing-value-{combo_no}-{inst_no}',
                      type='number', inputMode='numeric',
                      placeholder='-',
                      value=0.75,
                      min=0.5, max=2, step=0.1,
                      style={'width': '40%',
                             'float': 'Right',
                             'marginTop': 3,
                             'marginRight': 10,
                             'textAlign': 'Right'})
        ], style={'display': 'None'}),
        html.Div(style={'clear': 'both'}),
    ])

    inst_selection = html.Div([
        inst_presets,
        inst_name,
        html.Div(style={'clear': 'both'}),
        inst_wave,
        html.Div(style={'clear': 'both'}),
        inst_res,
        html.Div(style={'clear': 'both'}),
        inst_samp,
        html.Div(style={'clear': 'both'}),
        inst_snr,
    ])
    return inst_selection


def create_combination_selection(combo_no):
    combo_name = html.Div([
        html.H4('Instrument Combination: ',
                style={'display': 'inline-block'}),
        dcc.Input(id=f'comboname-{combo_no}',
                  value=f'Combo-{combo_no}',
                  style={'height': 25,
                         'width': 200,
                         'fontSize': 20,
                         'marginLeft': 20,
                         'display': 'inline-block'}),
    ], style={'display': 'block'})

    instruments = []
    for i in range(n_inst):
        if i == 0:
            disp = 'inline-block'
            button = 'Remove'
        else:
            disp = 'none'
            button = 'Add'
        instruments.append(
            html.Div([
                html.Div(id=f'instrument-selection-{combo_no}-{i}',
                         children=create_instrument_selection(combo_no, i),
                         style={'display': disp}),
                html.Button(id=f'inst-button-{combo_no}-{i}',
                            children=button,
                            style={'display': 'inline-block'}),

                ], style={'display': 'inline-block',
                          'verticalAlign': 'Top',
                          'width': '33%',
                          'height': '100%',
                          'borderRight': '1px solid gray',
                          'textAlign': 'Center'})
        )

    combo_selection = html.Div([
        combo_name,
        html.Div(instruments,
                 style={'margin': '0 auto',
                        'width': '99%'}),
    ], style={'display': 'block'})

    return combo_selection




###################
# Initialize Plots
###################
crlb_plot_layout = go.Layout(
                       yaxis=dict(range=[-3, 0],
                                  type='log',
                                  title=dict(text='[X/Fe]',
                                             font=dict(size=24)),
                                  ticks='inside',
                                  ticktext=np.concatenate([[0.001],
                                                           np.repeat("", 8),
                                                           [0.01],
                                                           np.repeat("", 8),
                                                           [0.1],
                                                           np.repeat("", 8),
                                                           [1.0]]),
                                  tickvals=np.concatenate([np.linspace(0.001, 0.009, 9),
                                                           np.linspace(0.01, 0.09, 9),
                                                           np.linspace(0.1, 1.0, 10)]),
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks'),

                       xaxis=dict(range=[-0.5, len(sample_labels)-0.5],
                                  ticks='inside',
                                  ticktext=sample_labels,
                                  tickvals=np.linspace(0, len(sample_labels), len(sample_labels)-1),
                                  tickangle=45,
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks'),
                       plot_bgcolor='white')

grad_plot_layout = go.Layout(width=1000, height=600,
                             yaxis=dict(title=dict(text='df/dX',
                                                   font=dict(size=24)),
                                        range=[-0.5, 0.5],
                                        ticks='inside',
                                        showline=True,
                                        linewidth=2,
                                        linecolor='black',
                                        gridcolor='rgba(0,0,0,0.25)',
                                        zeroline=False,
                                        mirror='ticks',
                                        domain=[0.5, 1.0]),
                             yaxis2=dict(title=dict(text='SNR / pixel',
                                                    font=dict(size=24)),
                                         range=[0.0, 100],
                                         ticks='inside',
                                         showline=True,
                                         linewidth=2,
                                         linecolor='black',
                                         gridcolor='rgba(0,0,0,0.25)',
                                         zeroline=False,
                                         mirror='ticks',
                                         domain=[0.0, 0.5]),
                             xaxis=dict(title=dict(text=u'Wavelength (\u212B)',
                                                   font=dict(size=24)),
                                        range=[3e3, 18e3],
                                        ticks='inside',
                                        tickangle=0,
                                        showline=True,
                                        linewidth=2,
                                        linecolor='black',
                                        gridcolor='rgba(0,0,0,0.25)',
                                        zeroline=False,
                                        mirror='ticks'),
                             xaxis2=dict(title=dict(text=u'Wavelength (\u212B)',
                                                   font=dict(size=24)),
                                        range=[3e3, 18e3],
                                        ticks='inside',
                                        tickangle=0,
                                        showline=True,
                                        linewidth=2,
                                        linecolor='black',
                                        gridcolor='rgba(0,0,0,0.25)',
                                        zeroline=False,
                                        mirror='ticks'),
                             plot_bgcolor='white'
                             )
grad_plot_data = [go.Scatter(x=[], y=[], yaxis='y1', xaxis='x1'),
                  go.Scatter(x=[], y=[], yaxis='y2', xaxis='x2'),
                  ]
fig = make_subplots(rows=2, cols=1, vertical_spacing=0, shared_xaxes=True)
fig.append_trace(go.Scatter(x=[], y=[]), row=1, col=1)
fig.append_trace(go.Scatter(x=[], y=[]), row=2, col=1)
fig.update_layout(width=1000, height=800,
                  margin=dict(l=75, r=50, t=30, b=50),
                  yaxis=dict(title=dict(text='df/dX',
                                        font=dict(size=24)),
                             range=[-0.5, 0.5],
                             ticks='inside',
                             showline=True,
                             linewidth=2,
                             linecolor='black',
                             gridcolor='rgba(0,0,0,0.25)',
                             zeroline=False,
                             mirror='ticks',),
                  yaxis2=dict(title=dict(text='SNR (pixel<sup>-1</sup>)',
                                         font=dict(size=24)),
                              range=[-0.5, 0.5],
                              ticks='inside',
                              showline=True,
                              linewidth=2,
                              linecolor='black',
                              gridcolor='rgba(0,0,0,0.25)',
                              zeroline=False,
                              mirror='ticks',),
                  xaxis=dict(#title=dict(text=u'Wavelength (\u212B)',
                             #           font=dict(size=24)),
                             range=[3e3, 18e3],
                             ticks='inside',
                             tickangle=0,
                             showline=True,
                             linewidth=2,
                             linecolor='black',
                             gridcolor='rgba(0,0,0,0.25)',
                             zeroline=False,
                             mirror='ticks'),
                  xaxis2=dict(title=dict(text=u'Wavelength (\u212B)',
                                         font=dict(size=24)),
                              range=[3e3, 18e3],
                              ticks='inside',
                              tickangle=0,
                              showline=True,
                              linewidth=2,
                              linecolor='black',
                              gridcolor='rgba(0,0,0,0.25)',
                              zeroline=False,
                              mirror='ticks'),
                  plot_bgcolor='white')


###################
# Initialize Layout
###################

# -----------
# Header
# -----------
header = html.Div([
    html.H1(children='Chem-I-Calc',
            style={'backgroundColor': UC_colors['blue'],
                   'color': UC_colors['white'],
                   'margin': 0,
                   'padding': 0}),
    ], style={'display': 'block',
              'width': '100%',
              'position': 'fixed',
              'left': 0,
              'top': 0,
              'zIndex': 10})

# -----------
# Star Select
# -----------
reference_star_select = html.Div([
    html.H4('Reference Star(s):',
            style={'display': 'block'}),
    dcc.Dropdown(id='reference-select',
                 options=[{'label': ref, 'value': ref} for ref in reference_stars],
                 placeholder='Select one or more reference stars',
                 multi=False)
])

# -----------
# Inst Select
# -----------
combinations = []
for i in range(n_combo):
    if i == 0:
        disp = 'block'
        button = 'Remove Instrument Combination'
    else:
        disp = 'none'
        button = 'Add Instrument Combination'
    combinations.append(
        html.Div(id=f'instrument-combination-{i}',
                 children=create_combination_selection(i),
                 style={'display': disp}),
    )
    combinations.append(
        html.Button(id=f'combo-button-{i}',
                    children=button,
                    style={'display': 'block',
                           'width': '50%',
                           'margin': 'auto',
                           'marginTop': 10}),
    )
    combinations.append(html.Hr())
combo_init = html.Div(combinations)

# -----------
# Results
# -----------
crlb_init = html.Div([
    # ............
    # Summary
    # ............
    html.Div([
        html.H4('Reference Star Summary:'),
        dash_table.DataTable(id='ref-table',
                             columns=[{'name': i, 'id': i} for i
                                      in ['name', 'Teff', 'logg', 'v_micro',
                                          '[Fe/H]', '[alpha/Fe]']],
                             sort_action='native',
                             sort_by=[{'column_id': 'name',
                                       'direction': 'asc'}],
                             style_as_list_view=True,
                             style_cell={'minWidth': f'{45/6}%', 'width': f'{45/6}%', 'maxWidth': f'{45/6}%'},
                             style_header={'backgroundColor': 'white',
                                           'fontWeight': 'bold'},
                             ),
        html.H4('Instrument Summary:'),
        dash_table.DataTable(id='inst-table',
                             columns=[{'name': i, 'id': i} for i
                                      in ['name', 'combo',
                                          'start', 'end',
                                          'R_res', 'R_samp']],
                             sort_action='native',
                             sort_by=[{'column_id': 'combo',
                                       'direction': 'asc'}],
                             style_as_list_view=True,
                             style_cell={'minWidth': f'{45/5}%', 'width': f'{45/5}%', 'maxWidth': f'{45/5}%'},
                             style_header={'backgroundColor': 'white',
                                           'fontWeight': 'bold'},
                             ),
    ], style={'display': 'inline-block',
              'width': '45%'}),
    # ............
    # Priors
    # ............
    html.Div([
        html.H4('Priors:',
                style={'paddingTop': '0px',
                       'height': '100%'}),
        dash_table.DataTable(id='prior-table',
                             columns=[{'name': i, 'id': i, 'editable': (i == 'Priors')}
                                      for i
                                      in ['Labels', 'Priors']],
                             data=[{'Labels': 'Teff', 'Priors': ''},
                                   {'Labels': 'log(g)', 'Priors': ''},
                                   {'Labels': '[Fe/H]', 'Priors': ''}],
                             tooltip_data=[{'Labels': '', 'Priors': prior_tooltips[i], 'type':'text'}
                                           for i in range(len(prior_tooltips))],
                             tooltip_delay=0,
                             style_as_list_view=True,
                             style_cell={'minWidth': '100px', 'width': '100px', 'maxWidth': '100px'},
                             style_header={'backgroundColor': 'white',
                                           'fontWeight': 'bold'},
                             )
    ], style={'display': 'inline-block',
              'width': '45%',
              'verticalAlign': 'Top',
              'paddingLeft': '50px'}),
    html.Div([
        html.Button(id='crlb-button',
                    disabled=True,
                    children='No instrument or star provided',
                    style={'width': '100%',
                           'color': '#ff8080',
                           'borderColor': '#ff8080'}),

    ], style={'display': 'block',
              'width': '50%',
              'margin': '30px auto'}),
    # ............
    # Results Tabs
    # ............
    dcc.Tabs(id='crlb-tabs', value='crlb',
             children=[
                 # CRLB Tab #
                 dcc.Tab(label='CRLB', value='crlb',
                         children=[
                             dcc.Graph(id='crlb-graph',
                                       style={'height': 500,
                                              'width': '100%'},
                                       figure=go.Figure(data=[], layout=crlb_plot_layout)),
                             html.Label(id='crlb-table-label',
                                        children='Cramer-Rao Lower Bounds:',
                                        style={'display': 'none'}),
                             dash_table.DataTable(id='crlb-table',
                                                  sort_action='native',
                                                  style_as_list_view=True),
                         ]),
                 # Grad Tab #
                 dcc.Tab(label='Gradients', value='grad',
                         children=[
                             html.Div('', style={'marginTop': 25}),
                             html.Label('Instruments:'),
                             dcc.Dropdown(id='grad-graph-inst',
                                          placeholder='Select instruments to plot',
                                          options=[],
                                          multi=True,
                                          ),
                             html.Label('Stellar Labels:'),
                             dcc.Dropdown(id='grad-graph-labels',
                                          placeholder='Select elements to plot',
                                          options=[{'label': lab, 'value': lab} for lab in lab_tab_text],
                                          multi=True,
                                          value=['Fe']
                                          ),
                             dcc.Graph(id='grad-graph',
                                       style={'height': '100%',
                                              'width': '100%'},
                                       #figure=go.Figure(data=grad_plot_data, layout=grad_plot_layout)),
                                       figure=fig),
                             html.Div('', style={'marginTop': 25}),

                         ]),
             ], style={'margin': 0,
                       'padding': 0}),
], style={'display': 'block'})




import numpy as np
import matplotlib.pyplot as plt
from itertools import cycle
import plotly.graph_objs as go
from plotly.subplots import make_subplots


def plotly_crlb(crlb):
    labs = crlb.index.values
    cols = crlb.columns
    c = plt.cm.get_cmap('viridis', len(cols))
    data = []
    for j, col in enumerate(cols):
        data.append(go.Scatter(
                        name=col,
                        x=crlb[col].loc[labs].index,
                        y=crlb[col].loc[labs].values,
                        marker={'size': 12,
                                'line': {'width': 2}},
                        mode='lines+markers',
                        line={'color': f'rgba{c(j)}'},
                        connectgaps=True))

    layout = go.Layout(legend=dict(x=0.99, y=0.01,
                                   xanchor='right',
                                   yanchor='bottom',
                                   font={'size': 16},
                                   orientation='h'),
                       yaxis=dict(range=[-3, 0],
                                  type='log',
                                  title=dict(text='[X/Fe]',
                                             font=dict(size=24)),
                                  ticks='inside',
                                  ticktext=np.concatenate([[0.001],
                                                           np.repeat("", 8),
                                                           [0.01],
                                                           np.repeat("", 8),
                                                           [0.1],
                                                           np.repeat("", 8),
                                                           [1.0]]),
                                  tickvals=np.concatenate([np.linspace(0.001, 0.009, 9),
                                                           np.linspace(0.01, 0.09, 9),
                                                           np.linspace(0.1, 1.0, 10)]),
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks'),
                       xaxis=dict(range=[-0.5, len(labs)-0.5],
                                  ticks='inside',
                                  tickangle=45,
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks'),
                       plot_bgcolor='white')
    return go.Figure(data=data, layout=layout)


def plotly_grads_snr(grad_dict, snr_dict, grad_names, snr_names, labels):
    c1 = plt.cm.get_cmap('viridis', len(labels)*len(grad_names))
    linecycler = cycle(['solid', 'dash', 'dashdot', 'dot'])

    fig = make_subplots(rows=2, cols=1, vertical_spacing=0, shared_xaxes=True)
    annotations = []
    shapes = []
    n = 0
    for i, lab in enumerate(labels):
        for j, name in enumerate(grad_names):
            fig.append_trace(go.Scatter(
                name=f'{name} ({lab})',
                x=grad_dict[name].index,
                y=grad_dict[name][lab],
                text=f'{name} ({lab})',
                hoverinfo='y+x+text',
                mode='lines',
                line={'color': f'rgba{c1(i+j)[:3] + (0.6,)}',
                      'dash': 'solid'}),
                row=1, col=1)
            snr_names2 = [snr_names[l] for l in range(len(snr_names)) if grad_names[j] in snr_names[l]]
            for k, snr_name in enumerate(snr_names2):
                print(snr_names2, snr_name)
                line = next(linecycler)
                fig.append_trace(go.Scatter(
                    name=snr_name,
                    x=snr_dict[snr_name].index,
                    y=snr_dict[snr_name].T.values[0],
                    showlegend=False,
                    text=snr_name,
                    hoverinfo='y+x+text',
                    mode='lines',
                    line={'color': f'rgba{c1(i+j)[:3] + (0.6,)}',
                          'dash': line}),
                    row=2, col=1)
                annotations.append(
                    go.layout.Annotation(
                        x=0.90, y=0.01+0.05*n,
                        showarrow=False,
                        text=snr_name,
                        font=dict(
                            color=f'rgba{c1(i+j)}',
                            size=16
                        ),
                        xanchor='right',
                        yanchor='bottom',
                        xref='paper', yref='paper')
                    )
                shapes.append(
                    go.layout.Shape(
                        type="line",
                        xref="paper",
                        yref="paper",
                        x0=0.92, x1=0.99,
                        y0=0.01 + 0.05 * k, y1=0.01 + 0.05*n,
                        line=dict(
                            color=f'rgba{c1(i+j)}',
                            dash=line,
                        )
                    )
                )
                n += 1

    fig.update_layout(annotations=annotations,
                      width=1000, height=800,
                      margin=dict(l=75, r=50, t=30, b=50),
                      legend=dict(x=0.99, y=0.51,
                                  xanchor='right',
                                  yanchor='bottom',
                                  font={'size': 16}),
                      yaxis=dict(title=dict(text='df/dX',
                                            font=dict(size=24)),
                                 ticks='inside',
                                 showline=True,
                                 linewidth=2,
                                 linecolor='black',
                                 gridcolor='rgba(0,0,0,0.25)',
                                 zeroline=False,
                                 mirror='ticks',
                                 hoverformat=".4f"),
                      yaxis2=dict(title=dict(text='SNR (pixel<sup>-1</sup>)',
                                             font=dict(size=24)),
                                  ticks='inside',
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks',
                                  hoverformat=".1f"),
                      xaxis=dict(
                          ticks='inside',
                          tickangle=0,
                          showline=True,
                          linewidth=2,
                          linecolor='black',
                          gridcolor='rgba(0,0,0,0.25)',
                          zeroline=False,
                          mirror='ticks',
                          hoverformat=".0f"),
                      xaxis2=dict(title=dict(text=u'Wavelength (\u212B)',
                                             font=dict(size=24)),
                                  ticks='inside',
                                  tickangle=0,
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks',
                                  hoverformat=".0f"),
                      plot_bgcolor='white')
    return fig



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from chemicalc.utils import generate_wavelength_template, load_filter_throughput, generate_tophat_throughput
from chemicalc.exception import InstError

filter_id = '1oANYac-Xc4n1TfTWXVjhfETHHBSTlUJP'
sampX = 3  # Placeholder for Unknown Wavelength Sampling


class InstConfig:
    def __init__(self, name: str, res: float, samp: float, start: float, end: float):
        self.name = name
        self.R_res = res
        self.R_samp = samp
        self.start_wavelength = start
        self.end_wavelength = end
        self.wave = generate_wavelength_template(self.start_wavelength, self.end_wavelength,
                                                 self.R_res * self.R_samp, truncate=False)
        self.snr = 100 * np.ones_like(self.wave)

    def set_wave(self):
        self.wave = generate_wavelength_template(self.start_wavelength, self.end_wavelength,
                                                 self.R_res * self.R_samp, truncate=False)

    def get_wave(self, truncate=False):
        wave = generate_wavelength_template(self.start_wavelength, self.end_wavelength,
                                            self.R_res*self.R_samp, truncate)
        self.wave = wave
        return wave

    def set_snr(self, snr_input):
        if (type(snr_input) == int) or (type(snr_input) == float) or (type(snr_input) == np.float64):
            self.snr = snr_input * np.ones_like(self.wave)
        elif (type(snr_input) == np.ndarray) and (snr_input.ndim == 2):
            snr_interpolator = interp1d(snr_input[0], snr_input[1],
                                        bounds_error=False, fill_value='extrapolate')
            self.snr = snr_interpolator(self.wave)
        elif (type(snr_input) == np.ndarray) and (snr_input.ndim == 1):
            fake_wave = np.linspace(self.wave.min(), self.wave.max(), snr_input.shape[0])
            snr_interpolator = interp1d(fake_wave, snr_input,
                                        bounds_error=False, fill_value='extrapolate')
            self.snr = snr_interpolator(self.wave)
        else:
            self.snr = snr_input.query_s2n(wavelength=self.wave)

        self.snr[self.snr < 0] = 0

    def summary(self):
        print(f'{self.name}\n' +
              f'{self.start_wavelength} < lambda (A) < {self.end_wavelength}\n' +
              f'R ~ {self.R_res}\n' +
              f'Sampling ~ {self.R_samp} pix/FWHM')


class DEIMOS(InstConfig):
    def __init__(self, name, res, samp, start, end):
        InstConfig.__init__(self, name, res, samp, start, end)
        self.facility = 'WMKO'
        self.instrument = 'DEIMOS'
        self.mode = 'Multi-Object Spectrograph'
        self.snr = []


class LRIS(InstConfig):
    def __init__(self, name, res, samp, start, end):
        InstConfig.__init__(self, name, res, samp, start, end)
        self.facility = 'WMKO'
        self.instrument = 'LRIS'
        self.mode = 'Multi-Object Spectrograph'


class HIRES(InstConfig):
    def __init__(self, name, res, samp, start, end):
        InstConfig.__init__(self, name, res, samp, start, end)
        self.facility = 'WMKO'
        self.instrument = 'HIRES'
        self.mode = 'Echelle Spectrograph'


class MIKE(InstConfig):
    def __init__(self, name, res, samp, start, end):
        InstConfig.__init__(self, name, res, samp, start, end)
        self.facility = 'Magellan'
        self.instrument = 'MIKE'
        self.mode = 'Echelle Spectrograph'


class M2FS(InstConfig):
    def __init__(self, name, res, samp, start, end):
        InstConfig.__init__(self, name, res, samp, start, end)
        self.facility = 'Magellan'
        self.instrument = 'M2FS'
        self.mode = 'Multi-Object Spectrograph'


class MiscInstrument(InstConfig):
    def __init__(self, name, res, samp, start, end, facility=None, instrument=None, mode=None):
        InstConfig.__init__(self, name, res, samp, start, end)
        self.facility = facility
        self.instrument = instrument
        self.mode = mode


class FilterSet:
    def __init__(self, name, filters=None):
        self.name = name
        filters_to_load = [fltr for fltr in filters if type(fltr) == str]
        custom_filters = [fltr for fltr in filters if type(fltr) == dict]
        invalid_filters = [fltr for fltr in filters if (type(fltr) != str) and (type(fltr) != dict)]

        throughput, wave_eff, fwhm = load_filter_throughput(filters=filters_to_load)
        for fltr in custom_filters:
            missing_inputs = [x for x in ['name', 'wave_eff', 'width', 'transmission'] if x not in list(fltr.keys())]
            if missing_inputs:
                print(f'WARNING! Filter {fltr} is missing the following inputs: "{", ".join(missing_inputs)}"')
            else:
                tmp_throughput, tmp_wave_eff, tmp_fwhm = generate_tophat_throughput(fltr['name'],
                                                                                    fltr['wave_eff'],
                                                                                    fltr['width'],
                                                                                    fltr['transmission'])
                throughput = pd.concat([throughput, tmp_throughput], axis=1)
                wave_eff = {**wave_eff, **tmp_wave_eff}
                fwhm = {**fwhm, **tmp_fwhm}
        for fltr in invalid_filters:
            print(f'WARNING! "{fltr}" is not a valid filter input')
        self.filters = filters_to_load + [fltr['name'] for fltr in custom_filters]
        self.throughput = throughput
        self.wave_eff = wave_eff
        self.fwhm = fwhm
        self.snr_dict = {fltr: 100 for fltr in self.filters}
        self.snr = np.array(list(self.snr_dict.values()))

    def set_snr(self, snr_input, quiet=False):
        if (type(snr_input) == int) or (type(snr_input) == float) or (type(snr_input) == np.float64):
            if snr_input > 0:
                self.snr_dict = {fltr: snr_input for fltr in self.filters}
            else:
                raise InstError('S/N must be > 0')
        elif type(snr_input) == dict:
            missing_fltr = [x for x in self.filters if x not in list(snr_input.keys())]
            extra_fltr = [x for x in snr_input.keys() if x not in list(self.filters)]
            if not quiet:
                for fltr in missing_fltr:
                    print(f'WARNING! "{fltr}" is not included in snr_input; S/N will remain {self.snr_dict[fltr]}.')
                for fltr in extra_fltr:
                    print(f'WARNING! "{fltr}" is not included in this FilterSet')
            for fltr in snr_input:
                if fltr in extra_fltr:
                    continue
                if snr_input[fltr] > 0:
                    self.snr_dict[fltr] = snr_input[fltr]
                else:
                    raise InstError('S/N must be > 0')
        elif (type(snr_input) == np.array) or (type(snr_input) == list):
            snr_input = np.array(snr_input)
            if len(snr_input) != len(self.filters):
                raise InstError('snr_input must be the same length as the number of filters')
            self.snr_dict = {self.filters[i]: snr_input[i] for i in range(len(snr_input))}
            self.snr = np.array(list(self.snr_dict.values()))
        else:
            raise InstError('snr_input must be a single value, a dictionary, a list, or a 1D numpy array')
        self.snr = np.array(list(self.snr_dict.values()))


class AllInstruments:
    def __init__(self):
        self.spectrographs = {  # # # WMKO # # #
                              'DEIMOS 1200G': DEIMOS('DEIMOS 1200G', res=6500, samp=4, start=6500, end=9000),
                              'DEIMOS 600ZD': DEIMOS('DEIMOS 600ZD', res=2100, samp=5, start=4000, end=9100),
                              'DEIMOS 900ZD': DEIMOS('DEIMOS 900ZD', res=2600, samp=5, start=4000, end=7500),
                              'DEIMOS 1200B': DEIMOS('DEIMOS 1200B', res=4000, samp=4, start=4000, end=6600),
                              'LRIS 600/4000 (b)': LRIS('LRIS 600/4000 (b)', res=1800, samp=4, start=3900, end=5500),
                              'LRIS 1200/7500 (r)': LRIS('LRIS 1200/7500 (r)', res=4000, samp=5, start=7700, end=9000),
                              'HIRESr 1.0"': HIRES('HIRESr 1.0"', res=35000, samp=3, start=3900, end=8350),
                              'HIRESr 0.8"': HIRES('HIRESr 0.8"', res=49000, samp=3, start=3900, end=8350),
                              'KCWI Small BL': MiscInstrument('KCWI Small BL', res=3600, samp=sampX, start=3500, end=5600), #dlambda = 2000
                              'KCWI Medium BL': MiscInstrument('KCWI Medium BL', res=1800, samp=sampX, start=3500, end=5600), #dlambda = 2000
                              'KCWI Large BL': MiscInstrument('KCWI Large BL', res=900, samp=sampX, start=3500, end=5600), #dlambda = 2000
                              'KCWI Small BM': MiscInstrument('KCWI Small BM', res=8000, samp=sampX, start=3500, end=5500), #dlambda = 800-900
                              'KCWI Medium BM': MiscInstrument('KCWI Medium BM', res=4000, samp=sampX, start=3500, end=5500), #dlambda = 800-900
                              'KCWI Large BM': MiscInstrument('KCWI Large BM', res=2000, samp=sampX, start=3500, end=5500), #dlambda = 800-900
                              'KCWI Small BH1': MiscInstrument('KCWI Small BH1', res=8000, samp=sampX, start=3500, end=4100), #dlambda = 400
                              'KCWI Medium BH1': MiscInstrument('KCWI Medium BH1', res=4000, samp=sampX, start=3500, end=4100), #dlambda = 400
                              'KCWI Large BH1': MiscInstrument('KCWI Large BH1', res=2000, samp=sampX, start=3500, end=4100), #dlambda = 400
                              'KCWI Small BH2': MiscInstrument('KCWI Small BH2', res=8000, samp=sampX, start=4000, end=4800), #dlambda = 370-440
                              'KCWI Medium BH2': MiscInstrument('KCWI Medium BH2', res=4000, samp=sampX, start=4000, end=4800), #dlambda = 370-440
                              'KCWI Large BH2': MiscInstrument('KCWI Large BH2', res=2000, samp=sampX, start=4000, end=4800), #dlambda = 370-440
                              'KCWI Small BH3': MiscInstrument('KCWI Small BH3', res=8000, samp=sampX, start=4700, end=5600), #dlambda = 470-530
                              'KCWI Medium BH3': MiscInstrument('KCWI Medium BH3', res=4000, samp=sampX, start=4700, end=5600), #dlambda = 470-530
                              'KCWI Large BH3': MiscInstrument('KCWI Large BH3', res=2000, samp=sampX, start=4700, end=5600), #dlambda = 470-530

                                # # # Magellan # # #
                              'MIKE 1" (r)': MIKE('MIKE 1" (r)', res=22000, samp=3, start=5000, end=10000),
                              'MIKE 1" (b)': MIKE('MIKE 1" (b)', res=28000, samp=4, start=3500, end=5000),
                              'M2FS MedRes': M2FS('M2FS MedRes', res=10000, samp=sampX, start=5100, end=5315),
                              'M2FS HiRes': M2FS('M2FS HiRes', res=18000, samp=sampX, start=5130, end=5185),

                                # # # MMT # # #
                              'Hectochelle': MiscInstrument('Hectochelle', res=20000, samp=6, start=5160, end=5280),
                              'Hectospec 270': MiscInstrument('Hectospec 270', res=1500, samp=5, start=3900, end=9200),
                              'Hectospec 600': MiscInstrument('Hectospec 600', res=5000, samp=5, start=5300, end=7800),
                              'Binospec 270': MiscInstrument('Binospec 270', res=1300, samp=4, start=3900, end=9200),
                              'Binospec 600a': MiscInstrument('Binospec 600a', res=2700, samp=3, start=4500, end=7000),
                              'Binospec 600b': MiscInstrument('Binospec 600b', res=3600, samp=3, start=6000, end=8500),
                              'Binospec 600c': MiscInstrument('Binospec 600c', res=4400, samp=3, start=7250, end=9750),
                              'Binospec 1000': MiscInstrument('Binospec 1000', res=3900, samp=3, start=3900, end=5400),

                                # # # VLT # # #
                              'MUSE': MiscInstrument('MUSE', res=2500, samp=sampX, start=4800, end=9300),
                              'XSHOOTER (UVB)': MiscInstrument('XSHOOTER (UVB)', res=6700, samp=5, start=3000, end=5500),
                              'XSHOOTER (VIS)': MiscInstrument('XSHOOTER (VIS)', res=11400, samp=4, start=5500, end=10200),
                              'XSHOOTER (NIR)': MiscInstrument('XSHOOTER (NIR)', res=5600, samp=4, start=10200, end=18000),
                              'GIRAFFE LR8': MiscInstrument('GIRAFFE LR8', res=6500, samp=sampX, start=8200, end=9400),
                              'GIRAFFE HR10': MiscInstrument('GIRAFFE HR10', res=19800, samp=sampX, start=5340, end=5620),
                              'GIRAFFE HR13': MiscInstrument('GIRAFFE HR13', res=22500, samp=sampX, start=6120, end=6400),
                              'GIRAFFE HR14A': MiscInstrument('GIRAFFE HR14A', res=28800, samp=sampX, start=6400, end=6620),
                              'GIRAFFE HR15': MiscInstrument('GIRAFFE HR15', res=19300, samp=sampX, start=6620, end=6960),

                                # # # LBT # # #
                              'MODS (b)': MiscInstrument('MODS (b)', res=1850, samp=4, start=3200, end=5500),
                              'MODS (r)': MiscInstrument('MODS (r)', res=2300, samp=4, start=5500, end=10500),

                                # # # JWST # # #
                              'NIRSpec G140M/F070LP': MiscInstrument('NIRSpec G140M/F070LP', res=1000, samp=sampX, start=7000, end=12700),
                              'NIRSpec G140M/F100LP': MiscInstrument('NIRSpec G140M/F100LP', res=1000, samp=sampX, start=9700, end=17999),
                              'NIRSpec G140H/F070LP': MiscInstrument('NIRSpec G140H/F070LP', res=2700, samp=sampX, start=7000, end=12700),
                              'NIRSpec G140H/F100LP': MiscInstrument('NIRSpec G140H/F100LP', res=2700, samp=sampX, start=9700, end=17999),

                                # # # MSE # # #
                              'MSE LR (b)': MiscInstrument('MSE LR (b)', res=3000, samp=sampX, start=3600, end=5400),
                              'MSE LR (g)': MiscInstrument('MSE LR (g)', res=3000, samp=sampX, start=5400, end=7200),
                              'MSE LR (r)': MiscInstrument('MSE LR (r)', res=3000, samp=sampX, start=7200, end=9500),
                              'MSE LR (NIR)': MiscInstrument('MSE LR (NIR)', res=3000, samp=sampX, start=9500, end=13000),
                              'MSE MR (b)': MiscInstrument('MSE MR (b)', res=5000, samp=sampX, start=3900, end=5000),
                              'MSE MR (g)': MiscInstrument('MSE MR (g)', res=5000, samp=sampX, start=5750, end=6900),
                              'MSE MR (r)': MiscInstrument('MSE MR (r)', res=5000, samp=sampX, start=7370, end=9000),

                              # # # MW Surveys # # #
                              'LAMOST': MiscInstrument('LAMOST', res=2000, samp=sampX, start=3900, end=9000),
                              'WEAVE': MiscInstrument('WEAVE', res=6000, samp=sampX, start=3700, end=10000),
                              'RAVE': MiscInstrument('RAVE', res=8000, samp=sampX, start=8400, end=8800),
                              'DESI (b)': MiscInstrument('DESI (b)', res=2500, samp=3, start=3600, end=5550),
                              'DESI (r)': MiscInstrument('DESI (r)', res=3500, samp=3, start=5550, end=6560),
                              'DESI (i)': MiscInstrument('DESI (i)', res=4500, samp=3, start=6560, end=9800),

                              'R100': MiscInstrument('R100', res=1e2, samp=sampX, start=3001, end=17999),
                              'R1000': MiscInstrument('R1000', res=1e3, samp=sampX, start=3001, end=17999),
                              'R10000': MiscInstrument('R10000', res=1e4, samp=sampX, start=3001, end=17999)
                              }



from pathlib import Path
import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
from mendeleev import element
from chemicalc.utils import data_dir, doppler_shift, convolve_spec, calc_MagAB, calc_gradient
from chemicalc.utils import download_package_files


precomputed_res = {'max': 300000,
                   #'high': 100000,
                   #'med': 50000,
                   #'low': 25000
                   }
precomputed_ref_id = {'max': '1I9GzorHm0KfqJ-wvZMVGbQDeyMwEu3n2'}
precomputed_cont_id = {'max': '1Fhx1KM8b6prtCGOZ3NazVeDQY-x9gOOU'}

label_file = data_dir.joinpath('reference_labels.h5')
label_id = '1-qCCjDXp2eNzRGCfIqI_2JZrzi22rFor'
if not label_file.exists():
    print('Downloading label_file')
    download_package_files(id=label_id,
                           destination=label_file)

reference_stars = list(pd.read_hdf(label_file, 'ref_list').values.flatten())
elements_included = [x.symbol for x in element(list(range(3, 100)))]
label_names = ['Teff', 'logg', 'v_micro'] + elements_included


class ReferenceSpectra:
    def __init__(self, reference: str, normalized=True, res='max', iron_scaled=False):
        self.reference = reference
        self.resolution = dict(init=precomputed_res[res])
        self.reference_file = data_dir.joinpath(f'reference_spectra_{self.resolution["init"]:06}.h5')
        if not self.reference_file.exists():
            print('Downloading reference file---this may take a few minutes but is only necessary once')
            download_package_files(id=precomputed_ref_id[res],
                                   destination=self.reference_file)
        wave_df = pd.read_hdf(self.reference_file, 'highres_wavelength')
        spec_df = pd.read_hdf(self.reference_file, reference)
        if not normalized:
            self.continuum_file = f'reference_continuum_{self.resolution["init"]:06}.h5'
            if not self.reference_file.exists():
                print('Downloading continuum file---this may take a few minutes but is only necessary once')
                download_package_files(id=precomputed_cont_id[res],
                                       destination=self.continuum_file)
            cont_df = pd.read_hdf(self.continuum_file, reference)
            spec_df *= cont_df
        label_df = pd.read_hdf(label_file, reference)
        label_df.index = label_names
        if not iron_scaled:
            label_df.loc[set(elements_included) ^ {'Fe'}] -= label_df.loc['Fe']

        self.wavelength = dict(init=wave_df.to_numpy().T[0])
        self.spectra = dict(init=spec_df.to_numpy().T)
        self.labels = label_df
        self.gradients = {}
        self.filters = {}

        self.nspectra = self.spectra['init'].shape[0]
        self.nlabels = self.labels.shape[0]

    def add_rv_spec(self, d_rv):
        self.labels.loc['RV'] = 0.0
        self.labels['fffff'] = self.labels['aaaaa']
        self.labels['ggggg'] = self.labels['aaaaa']
        self.labels.loc['RV', 'fffff'] += d_rv
        self.labels.loc['RV', 'ggggg'] -= d_rv
        tmp1 = doppler_shift(self.wavelength['init'], self.spectra['init'][0], d_rv)
        tmp2 = doppler_shift(self.wavelength['init'], self.spectra['init'][0], -d_rv)
        self.spectra['init'] = np.append(self.spectra['init'], tmp1[np.newaxis, :], axis=0)
        self.spectra['init'] = np.append(self.spectra['init'], tmp2[np.newaxis, :], axis=0)

    def convolve(self, instrument, name=None):
        if name is None:
            name = instrument.name
        outwave = instrument.get_wave()
        self.spectra[name] = convolve_spec(wave=self.wavelength['init'],
                                           spec=self.spectra['init'],
                                           resolution=instrument.R_res,
                                           outwave=outwave,
                                           res_in=self.resolution['init'])
        self.wavelength[name] = outwave
        self.resolution[name] = instrument.R_res

    def calc_synth_phot(self, filter_set, name=None, spectrum_name='init'):
        if name is None:
            name = filter_set.name
        if spectrum_name != 'init':
            filter_interp = interp1d(filter_set.throughput.index.values, filter_set.throughput.values,
                                     axis=0, fill_value=0, bounds_error=False)
            throughput = filter_interp(self.wavelength[spectrum_name])
        else:
            throughput = filter_set.throughput.values
        self.spectra[name] = calc_MagAB(f_nu=self.spectra[spectrum_name],
                                        throughput=throughput,
                                        wave=self.wavelength[spectrum_name])
        self.wavelength[name] = np.array(list(filter_set.wave_eff.values()))
        self.filters[name] = list(filter_set.throughput.columns)
        self.resolution[name] = 0

    def calc_gradient(self, name: str, symmetric: bool = True,
                      ref_included: bool = True, v_micro_scaling: float = 1,
                      d_rv: bool = None):
        self.gradients[name] = calc_gradient(self.wavelength[name], self.spectra[name], self.labels,
                                             symmetric=symmetric, ref_included=ref_included,
                                             v_micro_scaling=v_micro_scaling, d_rv=d_rv)
        self.gradients[name].columns = self.wavelength[name]

    def zero_gradients(self, name: str, labels: list):
        self.gradients[name].loc[labels] = 0

    def get_names(self):
        return list(self.spectra.keys())

    def reset(self):
        init_resolution = self.resolution['init']
        self.resolution.clear()
        self.resolution['init'] = init_resolution

        init_wavelength = self.wavelength['init']
        self.wavelength.clear()
        self.wavelength['init'] = init_wavelength

        init_spectra = self.spectra['init']
        self.spectra.clear()
        self.spectra['init'] = init_spectra

        del self.gradients



from flask import Flask
from dash import Dash

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

server = Flask('chemicalc_app')
app = Dash(server=server, external_stylesheets=external_stylesheets)

app.config['suppress_callback_exceptions'] = False


import numpy as np
from scipy.interpolate import interp1d
import mechanicalsoup
from chemicalc import exception as e

keck_options = {'instrument': ['lris', 'deimos', 'hires'],
                'mag type': ['Vega', 'AB'],
                'filter': ['sdss_r.dat', 'sdss_g.dat', 'sdss_i.dat', 'sdss_u.dat', 'sdss_z.dat',
                           'Buser_B.dat', 'Buser_V.dat', 'Cousins_R.dat', 'Cousins_I.dat'],
                'template': ['O5V_pickles_1.fits', 'B5V_pickles_6.fits',
                             'A0V_pickles_9.fits', 'A5V_pickles_12.fits',
                             'F5V_pickles_16.fits', 'G5V_pickles_27.fits',
                             'K0V_pickles_32.fits', 'K5V_pickles_36.fits',
                             'M5V_pickles_44.fits'],
                'grating (DEIMOS)': ['600Z', '900Z', '1200G', '1200B'],
                'grating (LRIS)': ['600/7500', '600/10000', '1200/9000', '400/8500', '831/8200'],
                'grism (LRIS)': ['B300', 'B600'],
                'binning (DEIMOS)': ['1x1'],
                'binning (LRIS)': ['1x1', '2x1', '2x2', '3x1'],
                'binning (HIRES)': ['1x1', '2x1', '2x2', '3x1'],
                'slitwidth (DEIMOS)': ['0.75', '1.0', '1.5'],
                'slitwidth (LRIS)': ['0.7', '1.0', '1.5'],
                'slitwidth (HIRES)': ["C5", "E4", "B2", "B5", "E5", "D3"],
                'slitwidth arcsec (HIRES)': [1.15, 0.40, 0.57, 0.86, 0.80, 1.72],
                'dichroic (LRIS)': ['D560'],
                'central wavelength (DEIMOS)': ['5000', '6000', '7000', '8000'],
                }


class Sig2NoiseWMKO:
    def __init__(self, instrument, exptime, mag, template,
                 magtype='Vega', band='Cousins_I.dat',
                 airmass=1.1, seeing=0.75, redshift=0):
        if instrument not in keck_options['instrument']:
            raise e.S2NInputError(f"{instrument} not one of {keck_options['instrument']}")
        if magtype not in keck_options['mag type']:
            raise e.S2NInputError(f"{magtype} not one of {keck_options['mag type']}")
        if band not in keck_options['filter']:
            raise e.S2NInputError(f"{band} not one of {keck_options['filter']}")
        if template not in keck_options['template']:
            raise e.S2NInputError(f"{template} not one of {keck_options['template']}")
        self.instrument = instrument
        self.mag = mag
        self.magtype = magtype
        self.filter = band
        self.template = template
        self.exptime = exptime
        self.airmass = airmass
        self.seeing = seeing
        self.redshift = redshift


class Sig2NoiseDEIMOS(Sig2NoiseWMKO):
    def __init__(self, grating, exptime, mag, template,
                 magtype='Vega', band='Cousins_I.dat',
                 cwave='7000', slitwidth='0.75', binning='1x1',
                 airmass=1.1, seeing=0.75, redshift=0):
        Sig2NoiseWMKO.__init__(self, 'deimos', exptime, mag, template, magtype, band, airmass, seeing, redshift)
        if grating not in keck_options['grating (DEIMOS)']:
            raise e.S2NInputError(f"{grating} not one of {keck_options['grating (DEIMOS)']}")
        if binning not in keck_options['binning (DEIMOS)']:
            raise e.S2NInputError(f"{binning} not one of {keck_options['binning (DEIMOS)']}")
        if slitwidth not in keck_options['slitwidth (DEIMOS)']:
            raise e.S2NInputError(f"{slitwidth} not one of {keck_options['slitwidth (DEIMOS)']}")
        if cwave not in keck_options['central wavelength (DEIMOS)']:
            raise e.S2NInputError(f"{cwave} not one of {keck_options['central wavelength (DEIMOS)']}")
        self.grating = grating
        self.binning = binning
        self.slitwidth = slitwidth
        self.cwave = cwave

    def query_s2n(self, wavelength='default'):
        url = 'http://etc.ucolick.org/web_s2n/deimos'
        browser = mechanicalsoup.StatefulBrowser()
        browser.open(url)
        form = browser.select_form()
        form['grating'] = self.grating
        form['cwave'] = self.cwave
        form['slitwidth'] = self.slitwidth
        form['binning'] = self.binning
        form['exptime'] = str(self.exptime)
        form['mag'] = str(self.mag)
        form['ffilter'] = self.filter
        if self.magtype.lower() == 'vega':
            form['mtype'] = '1'
        elif self.magtype.lower() == 'ab':
            form['mtype'] = '2'
        form['seeing'] = str(self.seeing)
        form['template'] = self.template
        form['airmass'] = str(self.airmass)
        form['redshift'] = str(self.redshift)
        data = browser.submit_selected().json()
        snr = np.array(data['s2n']).T
        if type(wavelength) == np.ndarray:
            snr_interpolator = interp1d(snr[0], snr[1])
            return snr_interpolator(wavelength)
        elif wavelength == 'default':
            return snr
        else:
            raise e.S2NInputError("Wavelength input not recognized")


class Sig2NoiseLRIS(Sig2NoiseWMKO):
    def __init__(self, grating, grism, exptime, mag, template,
                 magtype='Vega', band='Cousins_I.dat',
                 dichroic='D560', slitwidth='0.7', binning='1x1',
                 airmass=1.1, seeing=0.75, redshift=0):
        Sig2NoiseWMKO.__init__(self, 'lris', exptime, mag, template, magtype, band, airmass, seeing, redshift)
        if grating not in keck_options['grating (LRIS)']:
            raise e.S2NInputError(f"{grating} not one of {keck_options['grating (LRIS)']}")
        if grism not in keck_options['grism (LRIS)']:
            raise e.S2NInputError(f"{grism} not one of {keck_options['grism (LRIS)']}")
        if binning not in keck_options['binning (LRIS)']:
            raise e.S2NInputError(f"{binning} not one of {keck_options['binning (LRIS)']}")
        if slitwidth not in keck_options['slitwidth (LRIS)']:
            raise e.S2NInputError(f"{slitwidth} not one of {keck_options['slitwidth (LRIS)']}")
        if dichroic not in keck_options['dichroic (LRIS)']:
            raise e.S2NInputError(f"{dichroic} not one of {keck_options['dichroic (LRIS)']}")
        self.grating = grating
        self.grism = grism
        self.binning = binning
        self.slitwidth = slitwidth
        self.dichroic = dichroic

    def query_s2n(self, wavelength='default'):
        url = 'http://etc.ucolick.org/web_s2n/lris'
        browser = mechanicalsoup.StatefulBrowser()
        browser.open(url)
        form = browser.select_form()
        form['grating'] = self.grating
        form['grism'] = self.grism
        form['dichroic'] = self.dichroic
        form['slitwidth'] = self.slitwidth
        form['binning'] = self.binning
        form['exptime'] = str(self.exptime)
        form['mag'] = str(self.mag)
        form['ffilter'] = self.filter
        if self.magtype.lower() == 'vega':
            form['mtype'] = '1'
        elif self.magtype.lower() == 'ab':
            form['mtype'] = '2'
        form['seeing'] = str(self.seeing)
        form['template'] = self.template
        form['airmass'] = str(self.airmass)
        form['redshift'] = str(self.redshift)
        data = browser.submit_selected().json()
        snr = np.array(data['s2n']).T
        if type(wavelength) == np.ndarray:
            snr_interpolator = interp1d(snr[0], snr[1])
            return snr_interpolator(wavelength)
        elif wavelength == 'default':
            return snr
        else:
            raise e.S2NInputError("Wavelength input not recognized")


class Sig2NoiseHIRES(Sig2NoiseWMKO):
    def __init__(self, slitwidth, exptime, mag, template,
                 magtype='Vega', band='Cousins_I.dat', binning='1x1',
                 airmass=1.1, seeing=0.75, redshift=0):
        Sig2NoiseWMKO.__init__(self, 'hires', exptime, mag, template, magtype, band, airmass, seeing, redshift)
        if binning not in keck_options['binning (HIRES)']:
            raise e.S2NInputError(f"{binning} not one of {keck_options['binning (HIRES)']}")
        if slitwidth not in keck_options['slitwidth (HIRES)']:
            raise e.S2NInputError(f"{slitwidth} not one of {keck_options['slitwidth (HIRES)']}")
        self.binning = binning
        self.slitwidth = slitwidth

    def query_s2n(self, wavelength='default'):
        url = 'http://etc.ucolick.org/web_s2n/hires'
        browser = mechanicalsoup.StatefulBrowser()
        browser.open(url)
        form = browser.select_form()
        form['slitwidth'] = self.slitwidth
        form['binning'] = self.binning
        form['exptime'] = str(self.exptime)
        form['mag'] = str(self.mag)
        form['ffilter'] = self.filter
        if self.magtype.lower() == 'vega':
            form['mtype'] = '1'
        elif self.magtype.lower() == 'ab':
            form['mtype'] = '2'
        form['seeing'] = str(self.seeing)
        form['template'] = self.template
        form['airmass'] = str(self.airmass)
        form['redshift'] = str(self.redshift)
        data = browser.submit_selected().json()
        snr = np.array(data['s2n']).T
        if type(wavelength) == np.ndarray:
            snr_interpolator = interp1d(snr[0], snr[1])
            return snr_interpolator(wavelength)
        elif wavelength == 'default':
            return snr
        else:
            raise e.S2NInputError("Wavelength input not recognized")



import os
from pathlib import Path
import numpy as np
from chemicalc.instruments import AllInstruments

spectrographs = AllInstruments().spectrographs

tmp_data_dir = Path(os.path.dirname(__file__)).joinpath('/tmp/')
tmp_data_dir.mkdir(exist_ok=True)


def load_preset_specs(inst_name):
    preset = spectrographs[inst_name]
    inst = dict(
        name=inst_name,
        start_wavelength=preset.start_wavelength,
        end_wavelength=preset.end_wavelength,
        R_res=preset.R_res,
        R_samp=preset.R_samp
    )
    return inst


lab_tab_text = np.array(["Teff", "g", "v_micro",
                         "Li", "Be", "B", "C", "N", "O", "F", "Ne",
                         "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K",
                         "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni",
                         "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb",
                         "Sr", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd",
                         "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs",
                         "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd",
                         "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta",
                         "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb",
                         "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa",
                         "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf"])


prior_tooltips = [u'1-\u03C3 prior on effective temperature in K. '
                  u'Leave empty to include no prior information.'
                  u' Set to 0 to fix value.',
                  u'1-\u03C3 prior on log(g) in dex.'
                  u' Leave empty to include no prior information.'
                  u' Set to 0 to fix value.',
                  u'1-\u03C3 prior on [Fe/H] in dex.'
                  u' Leave empty to include no prior information.'
                  u' Set to 0 to fix value.']


sample_labels = ['T<sub>eff</sub> (100 K)', 'log(g)', 'v<sub>micro</sub> (km/s)',
                 "Li", "Be", "B", "C", "N", "O", "F", "Ne",
                 "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K"
                 ]

snr_options = ['Constant', 'From ETC']
etc_options = ['WMKO']


import os
from pathlib import Path
import requests
from tqdm import tqdm
import numpy as np
import pandas as pd
from numpy.fft import rfftfreq
from scipy.interpolate import interp1d
from chemicalc import exception as e

data_dir = Path(os.path.dirname(__file__)).joinpath('data')
data_dir.mkdir(exist_ok=True)
filter_file = Path(data_dir.joinpath('filters.h5'))
filter_id = '1oANYac-Xc4n1TfTWXVjhfETHHBSTlUJP'


def generate_wavelength_template(start_wavelength: float, end_wavelength: float,
                                 resolution: float, truncate: bool = False):
    """

    :param start_wavelength:
    :param end_wavelength:
    :param resolution:
    :param truncate:
    :return:
    """
    '''
    Generate wavelength array with fixed resolution.

    args:
    start_wavelength = minimum wavelength of spectra to include
    end_wavelength = maximum wavelength of spectra to include
    resolution = resolving power of instrument (R = lambda / delta lambda)
    truncate = Boolean, if true, drop final pixel for which lambda > end_wavelength

    returns = wavelength grid of given resolution between start and end wavelengths
    '''
    # TODO: generate_wavelength_template doc-string
    wavelength_template = [start_wavelength]
    wavelength_now = start_wavelength

    while wavelength_now < end_wavelength:
        wavelength_now += wavelength_now / resolution
        wavelength_template.append(wavelength_now)
    wavelength_template = np.array(wavelength_template)

    if truncate:
        wavelength_template = wavelength_template[:-1]
    return wavelength_template


def calc_f_nu(spectra, radius, dist=10):
    """
    radius: radius of star in solar radii
    dist: distance to star in pc
    """
    r_sun_cm = 6.957e10  # radius of the Sun in cm
    pc_cm = 3.086e18  # 1 parsec in cm
    f_nu0 = spectra * 4 * np.pi
    f_nu = f_nu0 * ((radius[:, np.newaxis] * r_sun_cm) / (dist * pc_cm)) ** 2
    return f_nu


def all_filters():
    tmp = pd.read_hdf(data_dir + 'filters.h5', 'fwhm')
    return list(tmp.columns)


def load_filter_throughput(filters):
    if not filter_file.exists():
        print('Downloading filters.h5')
        download_package_files(id=filter_id,
                               destination=filter_file)
    throughput = pd.read_hdf(filter_file, 'throughput')[filters]
    wave_eff = pd.read_hdf(filter_file, 'wave_eff', index='wave_eff')[filters]
    wave_eff = {fltr: wave_eff[fltr].values[0] for fltr in wave_eff}
    fwhm = pd.read_hdf(filter_file, 'fwhm')[filters]
    fwhm = {fltr: fwhm[fltr].values[0] for fltr in fwhm}
    return throughput, wave_eff, fwhm


def generate_tophat_throughput(name, wave_eff, width, transmission):
    wave = pd.read_hdf(data_dir.joinpath('reference_spectra_300000.h5'), 'highres_wavelength').values.flatten()
    throughput = transmission * pd.DataFrame((np.abs(wave - wave_eff) <= width / 2).astype(float),
                                             index=wave, columns=[name])
    wave_eff = {name: wave_eff}
    fwhm = {name: width}
    return throughput, wave_eff, fwhm


def calc_MagAB(f_nu, throughput, wave):
    c_aa_s = 2.998e18  # speed of light in angstroms/sec
    num = np.trapz(f_nu[:, :, np.newaxis]
                   * throughput[np.newaxis, :, :]
                   * wave[np.newaxis, :, np.newaxis],
                   x=wave, axis=1)
    den = np.trapz(throughput[:, :]
                   / wave[:, np.newaxis],
                   x=wave, axis=0)
    f = 1/c_aa_s * num[:, :]/den[np.newaxis, :]
    m_AB = -2.5 * np.log10(f) - 48.6
    return m_AB


def convolve_spec(wave, spec, resolution, outwave, res_in=None):
    """

    :param wave:
    :param spec:
    :param resolution:
    :param outwave:
    :param res_in:
    :return:
    """
    # TODO: convolve_spec doc-string
    sigma_to_fwhm = 2.355

    width = resolution * sigma_to_fwhm
    sigma_out = (resolution * sigma_to_fwhm) ** -1
    if res_in is None:
        sigma_in = 0
    else:
        sigma_in = (res_in * sigma_to_fwhm) ** -1

    # Trim Wavelength Range
    nsigma_pad = 20.0
    wlim = np.array([outwave.min(), outwave.max()])
    wlim *= (1 + nsigma_pad / width * np.array([-1, 1]))
    mask = (wave > wlim[0]) & (wave < wlim[1])
    wave = wave[mask]
    if spec.ndim == 1:
        spec = spec[mask]
    else:
        spec = spec[:, mask]

    # Make Convolution Grid
    wmin, wmax = wave.min(), wave.max()
    nwave = wave.shape[0]
    nwave_new = int(2 ** (np.ceil(np.log2(nwave))))
    lnwave_new = np.linspace(np.log(wmin), np.log(wmax), nwave_new)
    wave_new = np.exp(lnwave_new)
    fspec = interp1d(wave, spec,
                     bounds_error=False, fill_value='extrapolate')
    spec = fspec(wave_new)
    wave = wave_new

    # Convolve via FFT
    sigma = np.sqrt(sigma_out ** 2 - sigma_in ** 2)
    invres_grid = np.diff(np.log(wave))
    dx = np.median(invres_grid)
    ss = rfftfreq(nwave_new, d=dx)
    taper = np.exp(-2 * (np.pi ** 2) * (sigma ** 2) * (ss ** 2))
    spec_ff = np.fft.rfft(spec)
    ff_tapered = spec_ff * taper
    spec_conv = np.fft.irfft(ff_tapered)

    # Interpolate onto outwave
    fspec = interp1d(wave, spec_conv,
                     bounds_error=False, fill_value='extrapolate')
    return fspec(outwave)


def doppler_shift(wave, spec, rv):
    """

    :param wave:
    :param spec:
    :param rv:
    :return:
    """
    c = 2.99792458e5  # km/s
    doppler_factor = np.sqrt((1 - rv/c)/(1 + rv/c))
    new_wavelength = wave * doppler_factor
    return np.interp(new_wavelength, wave, spec)


def calc_gradient(wave, spectra, labels, symmetric=True, ref_included=True, v_micro_scaling=1, d_rv=None):
    """
    :param wave:
    :param spectra:
    :param labels:
    :param symmetric:
    :param ref_included:
    :param v_micro_scaling:
    :return:
    """
    # TODO: calc_gradient doc-string
    nspectra = spectra.shape[0]
    nlabels = labels.shape[0]
    if ref_included:
        skip = 1
    else:
        skip = 0
    if symmetric:
        if nspectra-skip != 2*nlabels:
            raise e.GradientError(f"nspectra({nspectra-skip}) != 2*nlabel({2*nlabels})"
                                  + "\nCannot perform symmetric gradient calculation")
        dx = np.diag(labels.iloc[:, 1::2].values - labels.iloc[:, 2::2].values).copy()
        grad = spectra[1::2] - spectra[2::2]
    else:
        if not ref_included:
            raise e.GradientError(f"Reference Spectra must be included at index 0"
                                  + "to calculate asymmetric gradients")
        if nspectra - 1 == nlabels:
            dx = np.diag(labels.iloc[:, 0].values - labels.iloc[:, 1:].values).copy()
            grad = spectra[0] - spectra[1:]
        elif nspectra - 1 == 2 * nlabels:
            dx = np.diag(labels.iloc[:, 0].values - labels.iloc[:, 1::2].values).copy()
            grad = spectra[0] - spectra[1::2]
        else:
            raise e.GradientError(f"nspectra({nspectra - 1}) != nlabel({nlabels})"
                                  + f"or 2*nlabel({2 * nlabels})"
                                  + "\nCannot perform asymmetric gradient calculation")
    dx[labels.index == 'Teff'] /= 100  # Scaling dX_Teff
    dx[labels.index == 'v_micro'] /= v_micro_scaling  # Scaling dX_v_micro
    dx[labels.index == 'rv'] /= 10
    dx[dx == 0] = -np.inf
    return pd.DataFrame(grad / dx[:, np.newaxis], index=labels.index)


def calc_crlb(reference, instruments, priors=None, output_fisher=False):
    """

    :param reference:
    :param instruments:
    :param priors:
    :param output_fisher:
    :return:
    """
    # TODO: calc_crlb doc-string
    if type(instruments) == list:
        grad_list = []
        snr_list = []
        for instrument in instruments:
            grad_list.append(reference.gradients[instrument.name].values)
            snr_list.append(instrument.snr)
        grad = np.concatenate(grad_list, axis=1)
        snr2 = np.diag(np.concatenate(snr_list, axis=0))**2
    else:
        grad = reference.gradients[instruments.name].values
        snr2 = np.diag(instruments.snr)**2

    fisher_mat = (grad.dot(snr2)).dot(grad.T)
    diag_val = (np.abs(np.diag(fisher_mat)) < 1.)
    fisher_mat[diag_val, :] = 0.
    fisher_mat[:, diag_val] = 0.
    fisher_mat[diag_val, diag_val] = 10. ** -6
    fisher_df = pd.DataFrame(fisher_mat, columns=reference.labels.index, index=reference.labels.index)

    if priors:
        for label in priors:
            prior = priors[label]
            if prior is None:
                continue
            if label == 'Teff':
                prior /= 100
            if prior == 0:
                fisher_df.loc[label, :] = 0
                fisher_df.loc[:, label] = 0
                fisher_df.loc[label, label] = 1e-6
            else:
                fisher_df.loc[label, label] += prior**(-2)

    crlb = pd.DataFrame(np.sqrt(np.diag(np.linalg.pinv(fisher_df))), index=reference.labels.index)

    if output_fisher:
        return crlb, fisher_df
    else:
        return crlb


def sort_crlb(crlb, cutoff, sort_by='default'):
    """

    :param crlb:
    :param cutoff:
    :param sort_by:
    :return:
    """
    crlb_temp = crlb[:3].copy()
    crlb[crlb > cutoff] = np.NaN
    crlb[:3] = crlb_temp

    if sort_by == 'default':
        sort_by_index = np.sum(pd.isna(crlb)).idxmin()
    else:
        if sort_by == list(crlb.columns):
            sort_by_index = sort_by
        else:
            assert False, f"{sort_by} not in CR_Gradients_File"

    valid_ele \
        = np.concatenate(
        [crlb.index[:3],
         crlb.index[3:][np.min(crlb[3:], axis=1) < cutoff]])
    valid_ele_sorted \
        = np.concatenate(
        [crlb.index[:3],
         crlb.loc[valid_ele][3:].sort_values(sort_by_index).index])

    crlb = crlb.loc[valid_ele_sorted]
    return crlb


def kpc_to_mu(d: float):
    """

    :param d: distance in kpc
    :return: distance modulus
    """
    return 5 * np.log10(d * 1e3 / 10)


def download_package_files(id, destination):
    def get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value

        return None

    def save_response_content(response, destination):
        chunk_size = 32768

        with open(destination, "wb") as f:
            for chunk in tqdm(response.iter_content(chunk_size)):
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)

    url = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(url, params={'id': id}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id': id, 'confirm': token}
        response = session.get(url, params=params, stream=True)

    save_response_content(response, destination)


def download_all_files(overwrite=True):
    from chemicalc.reference_spectra import label_file, label_id
    from chemicalc.reference_spectra import precomputed_res, precomputed_ref_id, precomputed_cont_id

    if filter_file.exists() and not overwrite:
        print(f'{filter_file} exists')
    else:
        print(f'Downloading {filter_file}')
        download_package_files(id=filter_id,
                               destination=filter_file)

    if label_file.exists() and not overwrite:
        print(f'{label_file} exists')
    else:
        print(f'Downloading {label_file}.h5')
        download_package_files(id=label_id,
                               destination=label_file)

    for res in precomputed_res:
        resolution = precomputed_res[res]
        reference_file = data_dir.joinpath(f'reference_spectra_{resolution:06}.h5')
        reference_id = precomputed_ref_id[res]
        if reference_file.exists() and not overwrite:
            print(f'{reference_file} exists')
        else:
            print(f'Downloading {reference_file}')
            download_package_files(id=reference_id,
                                   destination=reference_file)

    for res in precomputed_res:
        resolution = precomputed_res[res]
        continuum_file = data_dir.joinpath(f'reference_continuum_{resolution:06}.h5')
        continuum_id = precomputed_cont_id[res]
        if continuum_file.exists() and not overwrite:
            print(f'{continuum_file} exists')
        else:
            print(f'Downloading {continuum_file}')
            download_package_files(id=continuum_id,
                                   destination=continuum_file)



from .server import app, server

import dash_html_components as html
import dash_core_components as dcc

from .layout import header, reference_star_select, combo_init, crlb_init

app.index_string = '''
<!DOCTYPE html>
<html>
    <head>
        {%metas%}
        <title>Chem-I-Calc</title>
        {%favicon%}
        {%css%}
    </head>
    <body>
        <div>If the page fails to load, try refreshing.</div>
        {%app_entry%}
        <footer>
            {%config%}
            {%scripts%}
            {%renderer%}
        </footer>
    </body>
</html>
'''

app.layout = html.Div([
    header,
    html.Div('', style={'marginTop': 75}),
    html.Hr(),
    html.Details([
        html.Summary('1. Select Reference Star', style={'fontSize': 30}),
        reference_star_select
    ], open=True),
    html.Hr(),
    html.Details([
        html.Summary('2. Set Instrument Specifications', style={'fontSize': 30}),
        combo_init
    ], open=True),
    html.Hr(),
    html.Details([
        html.Summary('3. Calculate Cramer-Rao Bounds', style={'fontSize': 30}),
        crlb_init
    ], open=True),
    html.Hr(),
    html.Div('', style={'marginTop': 25}),

    # Storage Divs
    dcc.Store(id='ref-store', storage_type='memory'),
    dcc.Store(id='inst-store', storage_type='memory'),
    dcc.Store(id='crlb-store', storage_type='memory'),
    dcc.Store(id='snr-store', storage_type='memory'),
    html.Div(id='crlb-complete', children=0, style=dict(display='none')),
    html.Div(id='grad-key-div', children=[], style=dict(display='none')),
    html.Div(id='snr-key-div', children=[], style=dict(display='none')),
    html.Div(id='snr-missing', children=0, style=dict(display='none')),
])



import pandas as pd
import numpy as np

from .server import app
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from .layout import n_inst, n_combo
from .utils import load_preset_specs

from chemicalc.s2n import keck_options

for i in range(n_combo):
    @app.callback(
        [Output(f'combo-button-{i}', 'children'),
         Output(f'instrument-combination-{i}', 'style')],
        [Input(f'combo-button-{i}', 'n_clicks')],
        [State(f'combo-button-{i}', 'children'),
         State(f'combo-button-{i}', 'id')
         ])
    def add_remove_combinations(n_clicks, button, combo_id):
        combo_number = int(combo_id[-1])
        if not n_clicks:
            new_button = button
            if combo_number == 0:
                new_inst_style = {'display': 'block'}
            else:
                new_inst_style = {'display': 'None'}
        else:
            if button == 'Remove Instrument Combination':
                new_button = 'Add Instrument Combination'
                new_inst_style = {'display': 'None'}
            else:
                new_button = 'Remove Instrument Combination'
                new_inst_style = {'display': 'block'}
        return new_button, new_inst_style


    for j in range(n_inst):
        @app.callback(
            [Output(f'inst-button-{i}-{j}', 'children'),
             Output(f'instrument-selection-{i}-{j}', 'style')],
            [Input(f'inst-button-{i}-{j}', 'n_clicks')],
            [State(f'inst-button-{i}-{j}', 'children'),
             State(f'inst-button-{i}-{j}', 'id')
             ])
        def add_remove_instrument(n_clicks, button, inst_id):
            inst_number = int(inst_id[-1])
            if not n_clicks:
                new_button = button
                if inst_number == 0:
                    new_inst_style = {'display': 'inline-block'}
                else:
                    new_inst_style = {'display': 'None'}
            else:
                if button == 'Remove':
                    new_button = 'Add'
                    new_inst_style = {'display': 'None'}
                else:
                    new_button = 'Remove'
                    new_inst_style = {'display': 'inline-block'}
            return new_button, new_inst_style

        @app.callback(
            [Output(f'instname-{i}-{j}', 'value'),
             Output(f'wavemin-{i}-{j}', 'value'),
             Output(f'wavemax-{i}-{j}', 'value'),
             Output(f'res-{i}-{j}', 'value'),
             Output(f'samp-{i}-{j}', 'value'),
             ],
            [Input(f'preset-{i}-{j}', 'value')]
        )
        def load_preset(preset_name):
            if not preset_name:
                return '', '', '', '', ''
            if preset_name:
                inst = load_preset_specs(preset_name)
                return (inst['name'],
                        inst['start_wavelength'], inst['end_wavelength'],
                        inst['R_res'], inst['R_samp'])

        @app.callback(
            [Output(f'constant-snr-{i}-{j}', 'style'),
             Output(f'etc-dropdown-{i}-{j}', 'style'),
             Output(f'wmko-inst-{i}-{j}', 'style'),
             Output(f'wmko-grating-{i}-{j}', 'style'),
             Output(f'wmko-grism-{i}-{j}', 'style'),
             Output(f'wmko-slitwidth-{i}-{j}', 'style'),
             Output(f'wmko-dichroic-{i}-{j}', 'style'),
             Output(f'wmko-binning-{i}-{j}', 'style'),
             Output(f'wmko-cwave-{i}-{j}', 'style'),
             Output(f'wmko-exptime-{i}-{j}', 'style'),
             Output(f'wmko-mag-{i}-{j}', 'style'),
             Output(f'wmko-magtype-{i}-{j}', 'style'),
             Output(f'wmko-filter-{i}-{j}', 'style'),
             Output(f'wmko-star-{i}-{j}', 'style'),
             Output(f'wmko-airmass-{i}-{j}', 'style'),
             Output(f'wmko-seeing-{i}-{j}', 'style'),
             ],
            [Input(f'snr-type-{i}-{j}', 'value'),
             Input(f'etc-dropdown-{i}-{j}', 'value'),
             Input(f'wmko-inst-{i}-{j}', 'value')]
        )
        def create_snr_input(snr_type, etc, inst):
            basic_show = {'textAlign': 'Left',
                          'marginRight': 10,
                          'display': 'block'}
            basic_hide = {'textAlign': 'Left',
                          'marginRight': 10,
                          'display': 'None'}

            constant_snr_style = {'display': 'None'}
            etc_dropdown_style = basic_hide
            wmko_inst_style = basic_hide
            wmko_grating_style = basic_hide
            wmko_grism_style = basic_hide
            wmko_slitwidth_style = basic_hide
            wmko_dichroic_style = basic_hide
            wmko_binning_style = basic_hide
            wmko_cwave_style = basic_hide
            wmko_exptime_style = {'display': 'None'}
            wmko_mag_style = {'display': 'None'}
            wmko_magtype_style = basic_hide
            wmko_filter_style = basic_hide
            wmko_star_style = basic_hide
            wmko_airmass_style = {'display': 'None'}
            wmko_seeing_style = {'display': 'None'}
            if snr_type == "Constant":
                constant_snr_style = {'display': 'block'}
            elif snr_type == "From ETC":
                etc_dropdown_style = basic_show
                if etc == "WMKO":
                    wmko_inst_style = basic_show
                    if inst == 'deimos':
                        wmko_grating_style = basic_show
                        wmko_slitwidth_style = basic_show
                        wmko_binning_style = basic_show
                        wmko_cwave_style = basic_show
                        wmko_exptime_style = basic_show
                        wmko_mag_style = basic_show
                        wmko_magtype_style = basic_show
                        wmko_filter_style = basic_show
                        wmko_star_style = basic_show
                        wmko_airmass_style = basic_show
                        wmko_seeing_style = basic_show
                    elif inst == 'lris':
                        wmko_grating_style = basic_show
                        wmko_grism_style = basic_show
                        wmko_slitwidth_style = basic_show
                        wmko_dichroic_style = basic_show
                        wmko_binning_style = basic_show
                        wmko_exptime_style = basic_show
                        wmko_mag_style = basic_show
                        wmko_magtype_style = basic_show
                        wmko_filter_style = basic_show
                        wmko_star_style = basic_show
                        wmko_airmass_style = basic_show
                        wmko_seeing_style = basic_show
                    elif inst == 'hires':
                        wmko_slitwidth_style = basic_show
                        wmko_binning_style = basic_show
                        wmko_exptime_style = basic_show
                        wmko_mag_style = basic_show
                        wmko_magtype_style = basic_show
                        wmko_filter_style = basic_show
                        wmko_star_style = basic_show
                        wmko_airmass_style = basic_show
                        wmko_seeing_style = basic_show
            return (constant_snr_style, etc_dropdown_style, wmko_inst_style,
                    wmko_grating_style, wmko_grism_style, wmko_slitwidth_style,
                    wmko_dichroic_style, wmko_binning_style, wmko_cwave_style,
                    wmko_exptime_style, wmko_mag_style, wmko_magtype_style,
                    wmko_filter_style, wmko_star_style, wmko_airmass_style,
                    wmko_seeing_style)

        @app.callback(
            [Output(f'wmko-grating-{i}-{j}', 'options'),
             Output(f'wmko-grating-{i}-{j}', 'value'),
             Output(f'wmko-slitwidth-{i}-{j}', 'options'),
             Output(f'wmko-slitwidth-{i}-{j}', 'value'),
             Output(f'wmko-binning-{i}-{j}', 'options'),
             Output(f'wmko-binning-{i}-{j}', 'value'),
             ],
            [Input(f'wmko-inst-{i}-{j}', 'value')]
        )
        def set_wmko_inst_options(inst):
            if inst == 'deimos':
                grating_options = [{'label': opt, 'value': opt} for opt in keck_options['grating (DEIMOS)']]
                slitwidth_options = [{'label': opt, 'value': opt} for opt in keck_options['slitwidth (DEIMOS)']]
                binning_options = [{'label': opt, 'value': opt} for opt in keck_options['binning (DEIMOS)']]
            elif inst == 'lris':
                grating_options = [{'label': opt, 'value': opt} for opt in keck_options['grating (LRIS)']]
                slitwidth_options = [{'label': opt, 'value': opt} for opt in keck_options['slitwidth (LRIS)']]
                binning_options = [{'label': opt, 'value': opt} for opt in keck_options['binning (LRIS)']]
            elif inst == 'hires':
                grating_options = []
                slitwidth_options = [{'label': opt, 'value': opt} for opt in keck_options['slitwidth (HIRES)']]
                binning_options = [{'label': opt, 'value': opt} for opt in keck_options['binning (HIRES)']]
            else:
                grating_options, slitwidth_options, binning_options = [], [], []
            return grating_options, None, slitwidth_options, None, binning_options, None


inst_states = []
for i in range(n_combo):
    for j in range(n_inst):
        inst_states.append(Input(f'combo-button-{i}', 'children'))
        inst_states.append(Input(f'inst-button-{i}-{j}', 'children'))
        inst_states.append(Input(f'instname-{i}-{j}', 'value')),
        inst_states.append(Input(f'comboname-{i}', 'value')),
        inst_states.append(Input(f'wavemin-{i}-{j}', 'value')),
        inst_states.append(Input(f'wavemax-{i}-{j}', 'value')),
        inst_states.append(Input(f'res-{i}-{j}', 'value')),
        inst_states.append(Input(f'samp-{i}-{j}', 'value'))


@app.callback(
    Output('inst-store', 'data'),
    inst_states
)
def update_inst_store(*inputs):
    combo_active = [True if state[:6] == 'Remove'
                    else False
                    for state in inputs[0::8]]
    inst_active = [True if state[:6] == 'Remove'
                   else False
                   for state in inputs[1::8]]

    df = pd.DataFrame(index=inputs[2::8])
    df['name'] = inputs[2::8]
    df['combo'] = inputs[3::8]
    df['active'] = [True if combo_active[i] and inst_active[i]
                    else False
                    for i in range(n_combo*n_inst)]
    df['start'] = inputs[4::8]
    df['end'] = inputs[5::8]
    df['R_res'] = inputs[6::8]
    df['R_samp'] = inputs[7::8]
    active_df = df[df['active']]
    final_df = active_df.replace(to_replace=['None', ''], value=np.nan).dropna()

    return final_df.to_dict('records')


@app.callback(
    Output('inst-table', 'data'),
    [Input('inst-store', 'data')]
)
def update_inst_table(data):
    if data is None:
        raise PreventUpdate
    data = [dict(t) for t in {tuple(d.items()) for d in data}]
    return data


snr_states = [Input('inst-store', 'data')]
for i in range(n_combo):
    for j in range(n_inst):
        snr_states.append(Input(f'combo-button-{i}', 'children'))
        snr_states.append(Input(f'inst-button-{i}-{j}', 'children'))
        snr_states.append(Input(f'instname-{i}-{j}', 'value'))
        snr_states.append(Input(f'comboname-{i}', 'value'))
        snr_states.append(Input(f'snr-type-{i}-{j}', 'value'))
        snr_states.append(Input(f'snr-const-value-{i}-{j}', 'value'))
        snr_states.append(Input(f'etc-dropdown-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-inst-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-grating-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-grism-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-slitwidth-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-dichroic-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-binning-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-cwave-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-exptime-value-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-mag-value-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-magtype-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-filter-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-star-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-airmass-value-{i}-{j}', 'value'))
        snr_states.append(Input(f'wmko-seeing-value-{i}-{j}', 'value'))


@app.callback(
    [Output('snr-missing', 'children'),
     Output('snr-store', 'data')],
    snr_states
)
def check_snr_states(data, *inputs):
    if data is None:
        raise PreventUpdate
    combo_active = np.array([True if state[:6] == 'Remove'
                             else False
                             for state in inputs[0::21]])
    inst_active = np.array([True if state[:6] == 'Remove'
                            else False
                            for state in inputs[1::21]])
    active_instruments = combo_active * inst_active

    df = pd.DataFrame()
    df['name'] = inputs[2::21]
    df['combo'] = inputs[3::21]
    df['active'] = active_instruments
    df['type'] = inputs[4::21]
    df['const'] = inputs[5::21]
    df['etc'] = inputs[6::21]
    df['inst'] = inputs[7::21]
    df['grating'] = inputs[8::21]
    df['grism'] = inputs[9::21]
    df['slitwidth'] = inputs[10::21]
    df['dichroic'] = inputs[11::21]
    df['binning'] = inputs[12::21]
    df['cwave'] = inputs[13::21]
    df['exptime'] = inputs[14::21]
    df['mag'] = inputs[15::21]
    df['magtype'] = inputs[16::21]
    df['bandpass'] = inputs[17::21]
    df['star'] = inputs[18::21]
    df['airmass'] = inputs[19::21]
    df['seeing'] = inputs[20::21]
    active_df = df[df['active']]

    wmko_snr_requirements = ['slitwidth', 'binning', 'exptime', 'mag', 'magtype',
                             'bandpass', 'star', 'airmass', 'seeing']
    deimos_snr_requirements = wmko_snr_requirements + ['grating', 'cwave']
    lris_snr_requirements = wmko_snr_requirements + ['grating', 'grism', 'dichroic']
    hires_snr_requirements = wmko_snr_requirements
    missing_snr = 0
    for i in active_df.index:
        if active_df.loc[i, 'type'] == 'Constant':
            missing_snr += active_df['const'] is None
        elif active_df.loc[i, 'type'] == 'From ETC':
            if active_df.loc[i, 'inst'] == 'deimos':
                missing_snr += np.any(active_df.loc[i, deimos_snr_requirements].values == None)
            elif active_df.loc[i, 'inst'] == 'lris':
                missing_snr += np.any(active_df.loc[i, lris_snr_requirements].values == None)
            elif active_df.loc[i, 'inst'] == 'hires':
                missing_snr += np.any(active_df.loc[i, hires_snr_requirements].values == None)
            else:
                missing_snr += True
        else:
            missing_snr += True
    if missing_snr:
        snr_store = []
    else:
        snr_store = active_df.to_dict('records')
    return missing_snr, snr_store




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.ticker import StrMethodFormatter
import plotly.graph_objs as go


def plotly_crlb(crlb):
    labs = crlb.index.values
    cols = crlb.columns
    c = plt.cm.get_cmap('viridis', len(cols))
    data = []
    for j, col in enumerate(cols):
        data.append(go.Scatter(
                        name=col,
                        x=crlb[col].loc[labs].index,
                        y=crlb[col].loc[labs].values,
                        marker={'size': 12,
                                'line': {'width': 2}},
                        mode='lines+markers',
                        line={'color': f'rgba{c(j)}'},
                        connectgaps=True))

    layout = go.Layout(legend=dict(x=0.99, y=0.01,
                                   xanchor='right',
                                   yanchor='bottom',
                                   font={'size': 16},
                                   orientation='h'),
                       yaxis=dict(range=[-3, 0],
                                  type='log',
                                  title=dict(text='[X/Fe]',
                                             font=dict(size=24)),
                                  ticks='inside',
                                  ticktext=np.concatenate([[0.001],
                                                           np.repeat("", 8),
                                                           [0.01],
                                                           np.repeat("", 8),
                                                           [0.1],
                                                           np.repeat("", 8),
                                                           [1.0]]),
                                  tickvals=np.concatenate([np.linspace(0.001, 0.009, 9),
                                                           np.linspace(0.01, 0.09, 9),
                                                           np.linspace(0.1, 1.0, 10)]),
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks'),
                       xaxis=dict(range=[-0.5, len(labs)-0.5],
                                  ticks='inside',
                                  tickangle=45,
                                  showline=True,
                                  linewidth=2,
                                  linecolor='black',
                                  gridcolor='rgba(0,0,0,0.25)',
                                  zeroline=False,
                                  mirror='ticks'),
                       plot_bgcolor='white')
    return go.Figure(data=data, layout=layout)


def multiplot_crlb(crlb_list, **kwargs):
    if 'crlb_min' in kwargs.keys():
        ymin = kwargs['crlb_min']
    else:
        ymin = 9e-4
    if 'crlb_max' in kwargs.keys():
        ymax = kwargs['crlb_max']
    else:
        ymax = 3
    if 'ncol_legend' in kwargs.keys():
        ncol = kwargs['ncol_legend']
    else:
        ncol = 1

    if type(crlb_list) is not list:
        crlb_list = [crlb_list]

    lead_crlb = np.argmax([len(crlb.index) for crlb in crlb_list])
    all_labs = crlb_list[lead_crlb].index
    nlabs = all_labs.shape[0]
    cols = crlb_list[0].columns
    c = plt.cm.get_cmap('viridis', len(cols))

    fig = plt.figure(figsize=(25, len(crlb_list) * 10))
    gs = GridSpec(len(crlb_list), 1)
    gs.update(hspace=0.0)

    for i, crlb in enumerate(crlb_list):
        if i == 0:
            ax = plt.subplot(gs[i, 0])
        else:
            ax = plt.subplot(gs[i, 0], sharex=ax)
        labs = crlb.index
        for j, col in enumerate(cols):
            mask = pd.notnull(crlb[col].loc[labs].values)
            plt.plot(crlb[col].loc[labs].index[mask],
                     crlb[col].loc[labs].values[mask],
                     marker='o', markersize=16, markeredgewidth=2,
                     linestyle='-', linewidth=2,
                     color=c(j), markeredgecolor='k',
                     label=col)
        plt.grid(True, 'both', 'both', linewidth=0.5, alpha=0.5)
        ax.set_ylabel(r'$\sigma$[X/Fe]', size=36)
        ax.set_xlim(-0.5, nlabs - 0.5)
        ax.set_ylim(ymin, ymax)
        ax.set_yscale('log')
        plt.legend(fontsize=30, ncol=ncol, loc='lower right')
        ax.tick_params(axis='both', which='minor', length=5)
        ax.tick_params(axis='both', which='both', direction='in', labelsize=30)
        ax.yaxis.set_major_formatter(StrMethodFormatter("{x:.3f}"))
        if i == len(crlb_list) - 1:
            ax.tick_params(axis='x', which='major', rotation=-45, length=10)
        else:
            ax.tick_params(axis='x', length=10, labelsize=0)
        for label in ax.get_xticklabels():
            label.set_horizontalalignment('left')
        if 'labels' in kwargs.keys():
            plt.text(0.975, 0.95, s=kwargs['labels'][i], fontsize=30,
                     horizontalalignment='right',
                     verticalalignment='top',
                     transform=ax.transAxes,
                     bbox=dict(facecolor='white', edgecolor='black', pad=10.0))
    return fig


def overplot_crlb(crlb_list, **kwargs):
    if 'crlb_min' in kwargs.keys():
        ymin = kwargs['crlb_min']
    else:
        ymin = 9e-4
    if 'crlb_max' in kwargs.keys():
        ymax = kwargs['crlb_max']
    else:
        ymax = 3
    if 'ncol_legend' in kwargs.keys():
        ncol = kwargs['ncol_legend']
    else:
        ncol = 1

    lead_crlb = np.argmax([len(crlb.index) for crlb in crlb_list])
    labs = crlb_list[lead_crlb].index
    nlabs = labs.shape[0]
    cols = crlb_list[0].columns

    fig = plt.figure(figsize=(20, 10))
    ax = plt.subplot(111)
    c = plt.cm.get_cmap('viridis', len(cols))
    lines = ['-', '--', ':', '-.']

    for i, crlb in enumerate(crlb_list):
        for j, col in enumerate(cols):
            if i == 0:
                label = col
            else:
                label = '_nolegend_'
            mask = pd.notnull(crlb[col].loc[labs].values)
            plt.plot(crlb[col].loc[labs].index[mask],
                     crlb[col].loc[labs].values[mask],
                     marker='o', markersize=16, markeredgewidth=2,
                     linestyle=lines[i], linewidth=2,
                     color=c(j), markeredgecolor='k',
                     label=label)
        plt.grid(True, 'both', 'both', linewidth=0.5, alpha=0.5)
        plt.ylabel(r'$\sigma$[X/Fe]', size=36)
        plt.xlim(-0.5, nlabs - 0.5)
        plt.ylim(ymin, ymax)
        plt.yscale('log')
        plt.legend(fontsize=30, ncol=ncol, loc='lower right')
        plt.xticks(fontsize=30, rotation=-45)
        ax.yaxis.set_major_formatter(StrMethodFormatter("{x:.3f}"))
        for label in ax.get_xticklabels():
            label.set_horizontalalignment('left')
        plt.yticks(fontsize=30)
    return fig



from setuptools import setup

setup(
    name='Chem-I-Calc',
    version='0.1.3dev3',
    url='https://github.com/NathanSandford/Chem-I-Calc',
    packages=['chemicalc', 'chemicalc_app'],
    scripts=['chemicalc_app/run_chemicalc'],
    author='Nathan Sandford',
    author_email='nathan_sandford@berkeley.edu',
    license='MIT',
    long_description=open('README.md').read(),
    install_requires=[
        'tqdm',
        'requests',
        'dash',
        'flask',
        'mendeleev',
        'matplotlib',
        'plotly',
        'scipy',
        'mechanicalsoup',
        'pandas',
        'numpy',
        'tables',
    ],
)


from .server import app
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash import callback_context

from chemicalc import reference_spectra as ref
from chemicalc.instruments import InstConfig
from chemicalc import s2n
from chemicalc.utils import calc_crlb, sort_crlb
from .plot import plotly_crlb, plotly_grads_snr

from .utils import tmp_data_dir

import pandas as pd


@app.callback(
    [Output('crlb-button', 'disabled'),
     Output('crlb-button', 'children'),
     Output('crlb-button', 'style'),
     ],
    [Input('ref-store', 'data'),
     Input('inst-store', 'data'),
     Input('snr-missing', 'children'),
     Input('crlb-button', 'n_clicks'),
     Input('crlb-complete', 'children')
     ]
)
def check_crlb_button(ref_data, inst_data, snr_missing, n_clicks, flag):
    if ref_data and inst_data:
        inst_name_conflict = 0
        name_set = {tmp['name'] for tmp in inst_data}
        for name in name_set:
            config_set = {f"{tmp['start']}-{tmp['end']}_R{tmp['R_res']}({tmp['R_samp']})"
                          for tmp
                          in inst_data
                          if tmp['name'] == name}
            if len(config_set) != 1:
                inst_name_conflict += 1
        if inst_name_conflict:
            children = 'Instrument Name Conflict'
            style = {'width': '100%',
                     'color': '#ff8080',
                     'borderColor': '#ff8080'}
            disabled = True
        elif snr_missing:
            children = 'Instrument Missing SNR'
            style = {'width': '100%',
                     'color': '#ff8080',
                     'borderColor': '#ff8080'}
            disabled = True
        else:
            children = 'Calculate CRLB!'
            style = {'width': '100%',
                     'borderWidth': '3px',
                     'fontSize': 16,
                     'color': 'green',
                     'borderColor': 'green'}
            disabled = False
    elif ref_data and not inst_data:
        children = 'No instrument provided'
        style = {'width': '100%',
                 'color': '#ff8080',
                 'borderColor': '#ff8080'}
        disabled = True
    elif inst_data and not ref_data:
        children = 'No  star provided'
        style = {'width': '100%',
                 'color': '#ff8080',
                 'borderColor': '#ff8080'}
        disabled = True
    else:
        children = 'No instrument or star provided'
        style = {'width': '100%',
                 'color': '#ff8080',
                 'borderColor': '#ff8080'}
        disabled = True

    context = callback_context.triggered[0]['prop_id'].split('.')[0]
    if context == 'crlb-complete':
        if n_clicks:
            if flag == 1:
                children = 'Calculate CRLB!'
                style = {'width': '100%',
                         'borderWidth': '3px',
                         'fontSize': 16,
                         'color': 'green',
                         'borderColor': 'green'}
                disabled = False
        else:
            children = 'No instrument or star provided'
            style = {'width': '100%',
                     'color': '#ff8080',
                     'borderColor': '#ff8080'}
            disabled = True
    elif context == 'crlb-button':
        if n_clicks:
            children = 'Calculating CRLB...'
            style = {'width': '100%',
                     'borderWidth': '3px',
                     'fontSize': 16,
                     'color': 'green',
                     'borderColor': 'green'}
            disabled = True
        else:
            children = 'No instrument or star provided'
            style = {'width': '100%',
                     'color': '#ff8080',
                     'borderColor': '#ff8080'}
            disabled = True

    return disabled, children, style


@app.callback(
    [Output('crlb-store', 'data'),
     Output('crlb-complete', 'children'),
     Output('grad-key-div', 'children'),
     Output('snr-key-div', 'children')
     ],
    [Input('crlb-button', 'n_clicks')],
    [State('inst-store', 'data'),
     State('ref-store', 'data'),
     State('prior-table', 'data'),
     State('snr-store', 'data')
     ]
)
def calculate_crlb(n_clicks, inst_data, ref_data, prior_data, snr_data):
    if not n_clicks:
        raise PreventUpdate
    print('Calculating CRLBs')
    if all(tmp['Priors'] == '' for tmp in prior_data):
        priors = None
    else:
        priors = {}
        for i, label in enumerate(['Teff', 'logg', 'Fe']):
            prior = prior_data[i]['Priors']
            priors[label] = float(prior) if prior != '' else None
    snr_names = []
    # max_res = max([inst_dict['R_res'] * inst_dict['R_samp'] for inst_dict in inst_data])
    for i, ref_dict in enumerate(ref_data):
        ref_name = ref_dict['name']
        star = ref.ReferenceSpectra(reference=ref_name, res='high')
        if i == 0:
            crlb_df = pd.DataFrame(index=star.labels.index)
            combo_dict = {tmp['combo']: [] for tmp in inst_data}
        for j, inst_dict in enumerate(inst_data):
            inst = InstConfig(name=inst_dict['name'],
                              start=inst_dict['start'],
                              end=inst_dict['end'],
                              res=inst_dict['R_res'],
                              samp=inst_dict['R_samp'])
            snr_dict = snr_data[j]
            if snr_dict['type'] == "Constant":
                snr = snr_dict['const']
            elif snr_dict['type'] == 'From ETC':
                if snr_dict['etc'] == 'WMKO':
                    if snr_dict['inst'] == 'deimos':
                        snr = s2n.Sig2NoiseDEIMOS(grating=snr_dict['grating'], exptime=snr_dict['exptime'],
                                                  mag=snr_dict['mag'], magtype=snr_dict['magtype'],
                                                  band=snr_dict['bandpass'], template=snr_dict['star'],
                                                  cwave=snr_dict['cwave'], slitwidth=snr_dict['slitwidth'],
                                                  binning=snr_dict['binning'], airmass=snr_dict['airmass'],
                                                  seeing=snr_dict['seeing'], redshift=0)
                    elif snr_dict['inst'] == 'lris':
                        snr = s2n.Sig2NoiseLRIS(grating=snr_dict['grating'], grism=snr_dict['grism'],
                                                dichroic=snr_dict['dichroic'], exptime=snr_dict['exptime'],
                                                mag=snr_dict['mag'], magtype=snr_dict['magtype'],
                                                band=snr_dict['bandpass'], template=snr_dict['star'],
                                                slitwidth=snr_dict['slitwidth'], binning=snr_dict['binning'],
                                                airmass=snr_dict['airmass'], seeing=snr_dict['seeing'], redshift=0)
                    elif snr_dict['inst'] == 'hires':
                        snr = s2n.Sig2NoiseHIRES(slitwidth=snr_dict['slitwidth'], exptime=snr_dict['exptime'],
                                                 mag=snr_dict['mag'], magtype=snr_dict['magtype'],
                                                 band=snr_dict['bandpass'], template=snr_dict['star'],
                                                 binning=snr_dict['binning'], airmass=snr_dict['airmass'],
                                                 seeing=snr_dict['seeing'], redshift=0)
                    else:
                        snr = 0
                else:
                    snr = 0
            else:
                snr = 0
            inst.set_snr(snr)
            snr_df = pd.DataFrame(inst.snr, index=inst.wave)
            snr_df.to_hdf(tmp_data_dir.joinpath('tmp_snr.h5'),
                          key=f"{inst_dict['name']} ({inst_dict['combo']})", mode='a')
            snr_names.append(f"{inst_dict['name']} ({inst_dict['combo']})")

            combo_dict[inst_dict['combo']].append(inst)

            star.convolve(inst)
            star.calc_gradient(inst.name, symmetric=True, ref_included=True, v_micro_scaling=1)
        for combo in combo_dict:
            crlb_df[combo] = calc_crlb(star, combo_dict[combo], priors=priors).round(3)

    crlb_df = sort_crlb(crlb_df, cutoff=0.3, sort_by='default')
    crlb_df.replace(1000, 0, inplace=True)
    crlb_df['Labels'] = crlb_df.index

    grad_names = star.get_names()
    grad_names.remove('init')
    for name in grad_names:
        if name == 'init':
            continue
        star.gradients[name].T.to_hdf(tmp_data_dir.joinpath('tmp_grad.h5'), key=name, mode='a')

    return crlb_df.to_dict('records'), 1, grad_names, snr_names


@app.callback(
    [Output('crlb-table', 'data'),
     Output('crlb-table', 'columns'),
     Output('crlb-table-label', 'style')
     ],
    [Input('crlb-store', 'data')]
)
def update_crlb_table(data):
    if data is None:
        raise PreventUpdate
    new_columns = [{'name': 'Labels', 'id': 'Labels'}]
    new_columns += [{'name': i, 'id': i} for i
                    in list(data[0].keys())
                    if i != 'Labels']
    style = {'display': 'block'}
    return data, new_columns, style


@app.callback(
    Output('crlb-graph', 'figure'),
    [Input('crlb-store', 'data')]
)
def update_crlb_graph(data):
    if data is None:
        raise PreventUpdate
    crlb = pd.DataFrame(data)
    crlb.set_index('Labels', drop=True, inplace=True)
    crlb.index = ['T<sub>eff</sub> (100 K)', 'log(g)', 'v<sub>micro</sub> (km/s)'] + list(crlb.index[3:])
    fig = plotly_crlb(crlb)
    return fig


@app.callback(
    Output('grad-graph', 'figure'),
    [Input('grad-graph-inst', 'value'),
     Input('snr-key-div', 'children'),
     Input('grad-graph-labels', 'value')]
)
def update_grad_graph(grad_names, snr_names, labels):
    if not grad_names:
        raise PreventUpdate
    grad_dict = {}
    snr_dict = {}
    for name in grad_names:
        grad_dict[name] = pd.read_hdf(tmp_data_dir.joinpath('tmp_grad.h5'), key=name)
    for name in snr_names:
        snr_dict[name] = pd.read_hdf(tmp_data_dir.joinpath('tmp_snr.h5'), key=name)

    fig = plotly_grads_snr(grad_dict, snr_dict, grad_names, snr_names, list(labels))
    return fig


@app.callback(
    [Output('grad-graph-inst', 'options'),
     Output('grad-graph-inst', 'value')],
    [Input('grad-key-div', 'children')]
)
def update_grad_graph_inst(names):
    if not names:
        raise PreventUpdate
    return [{'label': name, 'value': name} for name in names], names



from .server import app
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from chemicalc.utils import data_dir

import pandas as pd


@app.callback(
    Output('ref-store', 'data'),
    [Input('reference-select', 'value')]
)
def update_ref_store(refs):
    name = []
    Teff = []
    logg = []
    v_micro = []
    FeH = []
    alphaFe = []
    if not refs:
        return pd.DataFrame().to_dict('records')
    elif type(refs) != list:
        refs = [refs]
    for reference in refs:
        label_df = pd.read_hdf(data_dir.joinpath('reference_labels.h5'), reference)
        name.append(reference)
        Teff.append(label_df.loc['Teff', 'aaaaa'])
        logg.append(label_df.loc['logg', 'aaaaa'])
        v_micro.append(label_df.loc['v_micro', 'aaaaa'])
        FeH.append(label_df.loc['Fe', 'aaaaa'])
        alphaFe.append(label_df.loc['Mg', 'aaaaa'] - label_df.loc['Fe', 'aaaaa'])
    df = pd.DataFrame()
    df['name'] = name
    df['Teff'] = Teff
    df['logg'] = logg
    df['v_micro'] = v_micro
    df['[Fe/H]'] = FeH
    df['[alpha/Fe]'] = alphaFe
    return df.to_dict('records')


@app.callback(
    Output('ref-table', 'data'),
    [Input('ref-store', 'data')]
)
def update_ref_table(data):
    if data is None:
        raise PreventUpdate
    return data
